/*  AX25 Level 2 interface  --  KE3Z 08/9/87  */

#include "buffer.h"
#include "iface.h"
#include "timer.h"
#include "ax25.h"
#include "ax25l2.h"
#include "netuser.h"
#include "data.h"

struct axcb *links; /* Main should init this  = NULL; */

extern unsigned char node_addr[16];
extern struct ax25_addr L3CALL[1], L3DIGI[1];

extern struct ax25_parms *l3parms;
int l3_st_up(), l2_window();
void level3(), NULLFCN();

/*  lookup_axcb -- searches through the list of axcbs for a match to the
 *  input parameters and returns a pointer to the first matching axcb.
 *  returns NULL if no match found.
 */
struct axcb *
lookup_axcb(laddr, raddrx, iface, alenx)
struct ax25_addr *laddr;	/* Local address to match */
struct ax25_addr raddrx[];	/* Remote address to match */
struct interface *iface;	/* Interface to match */
int alenx;
{
	register struct axcb *axcb;
	int callcmp();
	static int i, j;
	static struct ax25_addr *raddr;	/* Remote address to match */
	static int alen;

	alen=alenx;
	if ((raddr=raddrx) == NULL || laddr == NULL) return NULL;	
	for (axcb = links; axcb != NULL; axcb = axcb->next) {
	if (callcmp(laddr, &axcb->myaddr) &&	/* MYCALL match */
	   (iface == axcb->iface) &&		/* Interface matches */
	   (axcb->alen == alen) &&		/* Path same length */ 
	   (callcmp(raddr,axcb->path))) 	/* Src matches */ {
		j=alen;
		for (i=1;i<alen;i++) /* Check path */ {
			j--;
			if ((((raddr[j].ssid & (SSID | REPEATED)) ^ REPEATED)
			    != (axcb->path[i].ssid & (SSID | REPEATED)))
			    || !callcmp(&raddr[alen-i],&axcb->path[i]))
							goto no_match;
			/* The above insures the REPEATED bits are
				   right (the path in axcb has them toggled) */
		}
		return axcb;
		}
	no_match: continue;
	}
	return NULL;
}

/* ax25l2 -- handler of received AX.25 frames. */
struct datastr *
ax25l2(ifacex, bpx)		/* If it returns the frame then free it */
struct interface *ifacex;	/* Interface frame was received on */
struct datastr *bpx;		/* Received frame */
{
	static struct interface *iface;	/* Interface frame was received on */
	static struct datastr *bp;	/* Received frame */
	static int control;
	static struct ax25_addr adr[9];
	static struct axcb *axcb;
	static char rxversion;
	static unsigned char *chr, *chd, x121[8];
	static char iscmd, pf, frmrej;
	static char digi, swpos;
	static struct change *ch;
	static struct change *rrcmd[] = {rr_cmd, rr_cmdp};
	static struct change *rejcmd[] = {rej_cmd, rej_cmdp};
	static struct change *rnrcmd[] = {rnr_cmd, rnr_cmdp};
	static struct change *rrresp[] = {rr_resp, rr_respf};
	static struct change *rejresp[] = {rej_resp, rej_respf};
	static struct change *rnrresp[] = {rnr_resp, rnr_respf};
	static int i,j;
	static char numadr, numadrm;	/* Number of addresses (callsigns) */

	struct datastr *proci(), *setaddr();
	int callcmp();

	iface=ifacex;
	bp=bpx;
	iscmd=0;
	pf=0;
	frmrej=0;
	ch=NULL;

	/* Save Buffer start address incase this is a digipeat request */
	chr = bp->rdata;

	/* Move the address field into local storage */
	numadr=digi=swpos=0;
	if (buflen(bp) > AXALEN) {
		for (;numadr<7;numadr++) {
			for (i=0;i<ALEN;i++) {
				if (((j=bgetch(&bp)) == EOF) || (j & E)) {
					j= EOF;
					break;
				}
				adr[numadr].call[i]=j;
			}
			if (j == EOF) break;
			chd=bp->rdata;	/* Set the repeated bit in the frame*/
			if ((j=bgetch(&bp)) == EOF) break;
			adr[numadr].ssid=j;
			if (!swpos && callcmp(&adr[numadr],&L3CALL[0])) {
				if (( (numadr>1) && !(j & REPEATED))
					|| numadr<2) swpos=numadr+1;
			}
			if(!digi && numadr>1 && !(j & REPEATED)) {
				digi=numadr+1;	/* Digi Req? */
				*chd |= REPEATED; /* AFTER we copy it */
			}
			if (j & E) break;
		}
		if (j == EOF) numadr=0;
		else numadr++;
	}

	control = bgetch(&bp);

	puthex2("rx",control,"num",numadr);

	/* Check for valid address length, Ignore UI's and wrong PID */
/*	i = *bp->rdata; */

	if ((numadr > 7) || (numadr < 2)) return bp;

	if (numadr == 3) {
		if (digi) /* Not repeated yet */ {
			if (callcmp(&adr[2],&L3DIGI[0])) {
				bp->rdata = chr; /* Digi request, Restore addresses */
				(*iface->send)(iface,bp);
				return NULL;
			}
			else return bp; /* Not for me */
/*			else if (swpos == 1) return bp; /* To me */
		}
	}

	if (swpos == 2) return bp;		    /* I sent it */

	numadrm = numadr-1;
	if ((axcb=lookup_axcb(&adr[0],&adr[1],iface,numadrm)) == NULL) {
		if (!swpos) return bp; /* Not a switch request */
		if ((control & 0xef) != SABM) /* Not my link */ {
			goto send_dm;
		}

		if (callcmp(&adr[0],&adr[1])) goto send_dm;
					/* Can't connect to yourself! */

		if (numadr < 4) /* Connect request? */ {
			if (swpos != 1)	goto send_dm;	/* Not me! */
			if ((axcb =open_ax25(&adr[0],&adr[1],iface,numadrm))) {
				lnk_l3_pars(axcb);
			}
		}
		else /* Call request, set up an axcb to accept the call */ {
			if (swpos != digi) return bp;
			i=swpos+2;

			x121[0]=0;
			l3_adr(&adr[swpos],x121);
			if ((x121[0] == 4) && (i<=numadr)) {
				x121[0]=0;
				l3_adr(&adr[swpos+1],x121);
				if (x121[0]) i++;
			}

			if ((swpos > 4) || (i<numadr)) goto send_dm; /* DX */
		/*	if (i == numadr) adr[numadrm].ssid |= REPEATED;
					/* setpath toggles all repeated bits */
			axcb=open_ax25(&adr[0],&adr[1],iface,numadrm);
		}
	}

	if (axcb == NULL) goto send_dm;	/* No link block for this guy */

	/* Determine the protocol version of the received frame */
	if ((adr[0].ssid ^ adr[1].ssid) & C)
		rxversion = V2;
	else
		rxversion = V1;

	/* Determine if received frame is command or response */
	if ((adr[0].ssid & C) && !(adr[1].ssid & C)) iscmd = 1;
	if (control & PF) pf = 1;

	puthex2("cmd",iscmd,"pf",pf);

	/* Process the frame based on control field type
	 * I frames: process the receive-sequence number N(R), followed
	 * by the I field.
	 */

	if (!(control & 1)) {	/* I frame */
	 	axcb->window=l2_window(axcb->parms);
		procnr(axcb, control, iscmd);
		bp = proci(axcb, bp, control, iscmd);
	/* Reject any non-I frame with an I field except FRMR type */
	} else if ((control & ~PF) != FRMR && buflen(bp) != 0) {
		frmr(axcb, W+X, control, iscmd);
		return bp;

	/* S frames: process the N(R), then respond via state table
	 * nothing tricky here
	 */
	} else if ((control & 3) == S) { /* S frame */
	 	axcb->window=l2_window(axcb->parms);
		procnr(axcb, control, iscmd);
		switch (control & 0xf) {
		case RR:
			if (iscmd)
				ch = rrcmd[pf];
			else
				ch = rrresp[pf];
			break;
		case REJ:
			if (iscmd)
				ch = rejcmd[pf];
			else
				ch = rejresp[pf];
			break;
		case RNR:
			if (iscmd)
				ch = rnrcmd[pf];
			else
				ch = rnrresp[pf];
			break;
		default:	/* Invalid control field received */
			frmrej = 1;
			break;
		}
	} else 
	/* U frames: respond via state table except for SABM and UA,
	   which require a little special handling.
	 */
		switch (control & 0xef) { /* U frame */
		case SABM:
			axcb->iface = iface;
			setpath(axcb, &adr[1], 1, numadrm); /* Reset Path */
			link_reset(axcb); /* Reset the link */
			if (axcb->s_upcall) (void)(*axcb->s_upcall)(axcb,0,0);
			axcb->version = rxversion; /* Update the link version */
			idle_t1(axcb, 1); /* Start t3 running */
			ch = sabm_cmd;
			break;
		case DISC:
			ch = disc_cmd;
			break;
		case DM:
			ch = dm_resp;
			break;
		case UA:
			if (axcb->state == 2) /* Update the link version on connect */
				axcb->version = rxversion;
			ch = ua_resp;
			break;
		/* I suppose I should do something with the FRMR I field, but what? */
		case FRMR:
			ch = frmr_resp;
			break;
		default:	/* Invalid control field */
			frmrej = 1;
			break;
		}
	if (frmrej)		/* Reject bad control field */
		frmr(axcb, W, control, iscmd);
	else if (ch != NULL)	/* I frames handled elsewhere */
		respond(axcb, ch, control, iscmd);

	return bp; /* Free the frame, will be NULL if proci accepted data */

send_dm:	/* We get here if there is no link or against our rules */
	if ((control & 0xef) == DM) return bp;
	axcb=new_link();
	axcb->iface = iface;
	bcopy(adr[0].call, axcb->myaddr.call, ALEN);
	axcb->myaddr.ssid=adr[0].ssid;
	setpath(axcb, &adr[1], 1, numadrm);
	(*iface->send)(iface,setaddr(axcb,DM,0) );
	free(axcb);
	return bp;
}

/* respond -- act on the received frame via the passed state table
 */
respond(axcb, ch, rctl, rcmd)
register struct axcb *axcb;	/* Link */
struct change *ch;	/* State table pointer */
int rctl;		/* Received control field */
char rcmd;		/* Received frame was a command */
{
	static int st, acts;
	static int sctl;
	static int pf;

	puthex2("respond",ch,"ctl",rctl);

	pf=0;
	st = axcb->state -1;		/* State-table index */
	acts = ch[st].acts;		/* Get action flags */
	/* Reset the t1 timer and restore the retry count (n2) if specified */
	if (acts & RSTT1) idle_t1(axcb, 1);
	/* Here if a resulting frame is to be transmitted */
	sctl=ch[st].ctl;
	if (sctl == 255) sctl= -1; /* Just incase chars are unsigned/tm */
	if (sctl != -1) {
	/* FRMR handled specially */
		if (sctl == FRMR) {
			frmr(axcb, -1, 0, 0);
			return;
		}
	/* If I frame transmission specified, check that there is data on
	 * the transmit queue.  If not, change to S frame or nothing, based
	 * on the state of the RESPOND action flag.
	 */
		if (!(sctl & 1))
			if (axcb->iqueue == NULL)
				if (acts & RESPOND)
					sctl = (axcb->flags & BUSY) ? RNR : RR;
				else
					sctl = -1;
	/* Set P/F bit from received frame or action flag */
		if (axcb->version == V1 || rcmd) pf = rctl & PF;
		pf |= acts & PF;
	}
	/* Send the frame, change the state and all that dreck */
	actionl2(axcb, ch[st].newstate, sctl, pf, acts);
}

/* setpath -- update axcb path from received address field */
setpath(axcb, addr, swapdigi, alenx)
register struct axcb *axcb;		/* Link to update */
struct ax25_addr *addr;		/* received address field */
int swapdigi;			/* 1 if digis need to be reverse order */
int alenx;
{
	static int i, j, alen;

	alen = alenx;
	/* Store path in axcb */
	axcb->alen = alen;
	if (!alen) return;

	/* Clear all address-extension bits, and set the last */
	for (i = 0; i < alen; i++) {
		j=i;
		if (swapdigi && i>0) j=alen-i;
		bcopy(addr[i].call, axcb->path[j].call, ALEN);
		axcb->path[j].ssid = (addr[i].ssid & ~E );
		if (i>0) axcb->path[j].ssid ^= REPEATED; /* Toggle Repeated bit*/
	}
	if (i>1) axcb->path[alen-1].ssid |= E;	/* Set the last extension bit */
	else axcb->myaddr.ssid |= E;
}

/* link_reset -- stops the timers, clears the flags and resets the
 * sequence numbers for the link.
 */
link_reset(axcb)
register struct axcb *axcb;	/* Link */
{
	axcb->vs = axcb->ns = axcb->nr = 0;
	axcb->flags &= ~INVNS;
	kill(axcb->t1);
	kill(axcb->t2);
	kill(axcb->t3);
		/* Throw out any outstanding tx or rx data */
	free_queue(&axcb->iqueue);
	free_pkt(axcb->rcvq);
	axcb->icnt = 0;
	axcb->rcvcnt = 0;
}

/* axconnect -- initiate connect attempt on link (also known as a
 * "local start command" in the protocol spec). Returns 1 if OK, else 0
 */
axconnect(axcb)
register struct axcb *axcb;
{
	if (axcb->alen == 0) return 0;	/* No path set! */

	axcb->n2cnt = axcb->parms->retry;
	/* Send a SABM command, transition to state 2, start t1 */
	actionl2(axcb, 2, SABM, 0, CMD+RSTT1+SETT1);
	return 1;
}

void
setcalls(p,calls)
struct datastr *p;
register struct ax25_addr *calls;
{
	static int j;

	for (j=0;j<ALEN;j++) {
		bappch(p,calls->call[j]);
	}
	bappch(p,calls->ssid);
}

struct datastr *
setaddr(axcb,ctl,size)
register struct axcb *axcb;
int ctl,size;
{
	static struct datastr *p;
	static char i;

	/* Room for address and control and extra space needed by caller */
	p=new_buffer( 1 + size + (axcb->alen+1) * AXALEN);
	if (p==NULL) return NULL;

	setcalls(p,&axcb->path[0]);
	setcalls(p,&axcb->myaddr);
	for (i=1;i<axcb->alen;i++) setcalls(p,&axcb->path[i]);

	bappch(p,ctl);
	return p;
}

/* actionl2 -- takes action on the link.
 */
actionl2(axcb, newstatex, sctlx, pf, acts)
register struct axcb *axcb;	/* Link	*/
char newstatex;		/* New state, or 0 if no change	*/
int sctlx;		/* Control field to send */
int pf;			/* P/F bit needed */
char acts;		/* Action flags	*/
{
	static char oldstate, newstate;
	static int sctl;

	puthex2("actionl2",sctlx,"newst",newstate);

	newstate=newstatex;
	oldstate = axcb->state;		/* Save old state */
	/* Reset link if connecting */
	if ((sctl=sctlx) == SABM) {
		link_reset(axcb);
		axcb->version = axcb->parms->version;	/* Reset version */
	}
	/* State transition.  Set new state, then call state-change upcall if
	 * going into an information-transfer state (i.e. state > 4) from a
	 * setup state (i.e. state < 5)
	 */
	if (newstate != 0) {
		axcb->state = newstate;
		if (oldstate < 5 && newstate > 4) {
	/* Return busy indication if no upcall vector or upcall function
	 * returns error indication
	 */
			if (axcb->s_upcall != NULL &&
			 !(*axcb->s_upcall)(axcb, oldstate, newstate)) {
				sctl = DM;
				newstate = axcb->state = 1;
			} else {
	/* Now that state-change upcall has been made, perform transmit
	 * upcall if upcall function present and tx window is open
	 */
				if ((axcb->t_upcall != NULL) )
					(*axcb->t_upcall)(axcb);
			}
		}
	}

	/* Transmission handler.  Set C bits for command/response/none (V1) */

	if (sctl != -1 && !axcb->t2) /* MAY BE BUG 3/20 from 303 */ {
		axcb->myaddr.ssid &= ~C;
		axcb->path[0].ssid &= ~C;
		if (axcb->version >= V2)
			if (acts & CMD)
				axcb->path[0].ssid |= C;
			else
				axcb->myaddr.ssid |= C;

	/* If I frame, call the sending function and quit */
		if (!(sctl & 1) && (axcb->iqueue != NULL)) {
			sendi(axcb);
			return;
		}
	/* If S frame, merge receive-sequence variable into control field */

		if ((sctl & 3) == S) sctl |= (axcb->nr << 5);
		if (pf)	sctl |= PF;	/* Put P/F bit in */

	/* Call interface driver to transmit the frame */

		(*axcb->iface->send)(axcb->iface,setaddr(axcb,sctl,0));
	}
	/* If specified by action flag, start/restart t1 running */
	if (acts & SETT1) recovery(axcb);
	else sendi(axcb);		/* Send I frames if possible */

	/* If going disconnected, clear the link */
	if (axcb->state < 2) {
		link_reset(axcb);
		axcb->flags = 0;
	/* Inform state-change upcall function of disconnect */
		if (axcb->s_upcall != NULL && oldstate > 1)
			(*axcb->s_upcall)(axcb, oldstate, newstate);
	}
}

/* sendi - sends I frame(s).  Sends up to axcb->maxframe frames from
 * the transmit queue
 */

sendi(axcb)
register struct axcb *axcb;	/* Link */
{
	static struct datastr *bp;
	static struct datastr *dp;
	static int i;
	static char ctl;

	if (axcb->iqueue == NULL) return;	/* Nothing to send */
	if ((axcb->t1 != NULL) || (axcb->t2 != NULL) || (axcb->state<5)) return;
	/* Can't send I frames if t1 or t2 running or not connected */

	axcb->vs = axcb->ns;		/* Init send variable */
	bp = axcb->iqueue;		/* Point to top of queue */
	axcb->myaddr.ssid &= ~C; /* I frame is never a response */
	if (axcb->version == V1)
		axcb->path[0].ssid &= ~C;
	else				/* Make it a command */
		axcb->path[0].ssid |= C;
	/* Frame-send loop */
	for (i = 0; i < axcb->parms->maxframe; i++, bp = bp->next) {
		puthex2("txfrm",axcb->parms->maxframe,"",bp);
		if (bp == NULL)	break;		/* Quit at end of queue */
		ctl=(axcb->nr << 5) | (axcb->vs << 1); /* Form control field */
	/* Make a copy of the data, we might have to send it again */
		dp = dup_pkt(bp);
		if ((axcb->iface->send == NULL)	|| !dp) break;	/* Can't Send */
		dp= binsert( setaddr(axcb,ctl,0), dp );
		(*axcb->iface->send)(axcb->iface,dp);
		axcb->vs = (axcb->vs + 1) & 7;	/* Increment send variable */
	}
	recovery(axcb);		/* Start the t1 timer */
}

/* procnr -- process the N(R) value of a received I or S frame */
procnr(axcb, control, iscmd)
register struct axcb *axcb;	/* Link	*/
int control;			/* Received control field */
char iscmd;			/* 1 if received frame is a command */
{
	static char rs, top, acked;
	static struct datastr *bp;

	acked=0;
	/* Don't bother if we are'nt connected */
	if (axcb->state < 5)
		return;
	/* Break the sequence number out of the control field */
	rs = (control >> 5) & 7;
	/* Determine the proper highest and lowest acceptible sequence values
	 * If the N(S) value (lowest outstanding frame number) is greater than
	 * or equal to V(S) (highest outstanding frame number), adjust V(S) to
	 * a value greater than N(S) by adding 8 to it and to R(S).  (This
	 * happens when the modulo-8 sequence number "rolls over" to 0 during
	 * transmission.)
	 */
	if (axcb->vs >= axcb->ns)
		top = axcb->vs;
	else {
		top = axcb->vs + 8;
		if (rs < axcb->ns)
			rs += 8;
	}
	/* If R(S) isn't within the acceptible range, reject the frame */
	if (rs < axcb->ns || rs > top) {
		frmr(axcb, Z, control, iscmd);
		return;
	}
	/* For each acknowledged frame, increment N(S), yank the acked frame
	 * off the tx queue, and reduce the queue data count.
	 */
	while (rs > axcb->ns) {
		axcb->ns++;
		if (axcb->iqueue != NULL) {
			bp = axcb->iqueue;
			axcb->icnt -= buflen(bp);
			axcb->iqueue = bp->next;
			free_pkt(bp);
		}
	/* If a frame (or frames) is being acked, perform state-change stuff */
		acked = 1;
	}
	/* Normalize N(S) modulo 8 */
	axcb->ns &= 7;
	/* If frames were acked and the tx window is open, notify the transmit
	 * upcall handler
	 */
	 if (acked || axcb->window > axcb->icnt) {
		respond(axcb, (axcb->version == V1) ? val_nsv1 : val_ns, 0, 0);
		if (axcb->t_upcall != NULL) (*axcb->t_upcall)(axcb);
	}
}

/* proci -- process the send-sequence number and I field of received
 * frames
 */
struct datastr *
proci(axcb, bpx, controlx, iscmd)
register struct axcb *axcb;	/* Link */
struct datastr *bpx;	/* Received frame */
int controlx;		/* Received control field */
char iscmd;		/* 1 if received frame was a command */
{
	static int pid, c;
	static int control;
	static char vs;
	static struct datastr *bp1, *bp;
	void t2_timeout();

	bp=bpx;
	control=controlx;

	/* Don't bother if not connected */
	if (axcb->state < 5) return bp;

	/* If I field is too big, reject the frame */
	/* allow upto 4 extra bytes, PID and a trim 3 bytes for X.25 level 3 */
	if (buflen(bp) > axcb->parms->framesize+4) {
		frmr(axcb, Y, control, iscmd);
		return bp;
	}
	/* Ignore received frames if we are busy (rx queue full) */
	if (!(axcb->flags & BUSY)) {
	/* Get the PID field */
		if ( (pid=bgetch(&bp)) == -1)
			frmr(axcb, W+Y, control, iscmd);
	/* Break out the N(S) value and reset the "invalid N(S)" flag */
		vs = (control >> 1) & 7;
		axcb->flags &= ~INVNS;
	/* If this frame is not the expected one, set "invalid N(S)" */
		if (vs != axcb->nr)
			axcb->flags |= INVNS;
	/* Save control field for delayed response */
		axcb->rctl = control;
	/* If this is expected frame, accept it */
		if (vs == axcb->nr) {
			axcb->rpid = pid; /* Update received PID */
			axcb->nr = (axcb->nr + 1) & 7;	/* Increment our N(R) */
			if ((c=buflen(bp)) > 0) { /* Process data in frame */
				bp1 = axcb->rcvq;
				/* Update receive queue count */
				axcb->rcvcnt += c;
				/* Update window Size based on free memory */
				axcb->window=l2_window(axcb->parms);
	/* If receive window is now closed, perform go-busy routine */
				if (axcb->rcvcnt >= axcb->window) {
					axcb->flags |= BUSY;
					c = go_busy[axcb->state-1].newstate;
						/* Enter busy state */
					if (c != 0)
						axcb->state = c;
				}
/* this may be converting frames in to a data stream, should be done in the
   user module not a level 2 function!!! [append, next line] w2vy 10/86 */
				axcb->rcvq=binsert(axcb->rcvq, bp); 
				/* Put data on the rx queue */
				bp = NULL; /* Buffer isn't ours any more */
			/* Always pass a revd frame to the upper layer w2vy */
				if (axcb->r_upcall) (*axcb->r_upcall)(axcb);
			}
			reset_t3(axcb);	/* Link activity; reset t3 */
		}
	}
	/* If delayed-response is in effect, start t2 timer running */
/*	kill(axcb->t2);	/* Incase it was running */
	queue(&axcb->t2,t2_timeout,axcb->parms->resptime,axcb);

	return bp;
}
/* l2_window - Return what window size should be based on how much
/* memory the switch has left, max size is maxframe*framesize */
int
l2_window(parms)
register struct ax25_parms *parms;
{
	static int i;

	i=parms->maxframe * parms->framesize;
	switch (spclev()) {
		case FINE: return i;		/* Whole window */
		case LIMITED: return (i>>1);	/* Only 1/2 */
		case CRITICAL: return (0);	/* Nothing! */
	}
}

/* frmr -- reject bad received frame
 * "errtyp" should be -1 to repeat the previous FRMR frame
 */
frmr(axcb, errtyp, control, iscmd)
register struct axcb *axcb;	/* Link */
int errtyp;		/* Error type for FRMR I field */
int control;		/* Received control field */
char iscmd;		/* 1 if received frame is command */
{
	static struct datastr *bp;
/*	static char info[3]; */
	static char *ch;
	static char newstate;

	ch = axcb->info;
	newstate=0;
	if (errtyp != -1) {		/* Form new FRMR I field */
		newstate = 3;			/* Go to frame reject state */
	/* Form FRMR I field */
		*ch++ = errtyp;
		*ch = 0;
		if (iscmd) *ch |= PF;
		*ch++ |= (axcb->nr << 5) + (axcb->ns << 1);
		*ch = control;
		ch = axcb->info;
	}
	/* Make an datastr of the FRMR data and send it */
	bp = setaddr(axcb,FRMR,3);
	bappch(bp,*ch++);
	bappch(bp,*ch++);
	bappch(bp,*ch);
	(*axcb->iface->send)(axcb->iface,bp);
	actionl2(axcb, newstate, -1, 0, 0);	/* Transition to FRMR state */
}

/* t1_timeout -- t1 timer handler.  On timeout:
 * 1) Start t3 timer if not V1 protocol
 * 2) Decrement retry (N2) count if not 0
 * 3) If N2 decremented to 0, respond via "N2 exceeded" state table,
 *    otherwise, respond via "t1 expired" state table
 */
void
t1_timeout(axcb)
register struct axcb *axcb;	/* Link	*/
{
	idle_t1(axcb, 0);
	if (axcb->n2cnt)
		if (--axcb->n2cnt == 0) {
			respond(axcb, n2_exceeded, 0, 0);
			return;
		}
	respond(axcb, (axcb->version == V1) ? t1_expv1 : t1_exp, 0, 0);
}

/* t2_timeout -- t2 timer handler.  On timeout, if link "invalid N(S)
 * received" flag is set, respond via "invalid N(S)" state table, otherwise
 * respond via "i frame" state table.
 */
void
t2_timeout(axcb)
register struct axcb *axcb;	/* Link	*/
{
	if (axcb->r_upcall) (*axcb->r_upcall)(axcb);
	if (axcb->flags & INVNS)
		respond(axcb,(axcb->version == V1) ? inv_nsv1 : inv_ns,
					axcb->rctl, axcb->rcmd);
	else
		respond(axcb, (axcb->rctl & PF) ? i_cmdp : i_cmd,
					axcb->rctl, axcb->rcmd);
}

/* t3_timeout -- t3 timer handler.  On timeout, respond via "t3 expired"
 * state table.
 */
void
t3_timeout(axcb)
register struct axcb *axcb;	/* Link	*/
{
	respond(axcb, t3_exp, 0, 0);
}

/* recovery -- initializes (starts) t1 timer and stops t3 timer.
 * t1 timer period is calculated by:
 * p = (2*n + 1) * f
 * where p=period in seconds, n = number of digi addresses in the link
 * path, and f=frack
 */
/* static .. what is a static function?? /tm */
recovery(axcb)
register struct axcb *axcb;	/* Link */
{
	static int i;

	kill(axcb->t3);
	i=axcb->alen;
	if (i>2) i -= 2;	/* Subtract off the switch call and address */
	i = ( 2 *(i-1) +1) * (axcb->parms->frack *
		(axcb->parms->retry - axcb->n2cnt + 1)) + rnd(3);
	queue(&axcb->t1,t1_timeout,i,axcb);
}

/* idle_t1 -- stops the t1 timer, optionally resetting the retry (N2) count.
 * starts the inactive link (t3) timer if the link isn't using V1 protocol
 */
idle_t1(axcb, reset)
register struct axcb *axcb;	/* Link	*/
int reset;		/* 1 to reset retry count */
{
	kill(axcb->t1);
	if (reset)
		axcb->n2cnt = axcb->parms->retry;
	reset_t3(axcb);
}

/* reset_t3 - restart t3 (link activity occurred) */
reset_t3(axcb)
register struct axcb *axcb;	/* Link	*/
{
	if (axcb->version != V1) {
/*		kill(axcb->t3); */
		queue(&axcb->t3,t3_timeout,axcb->parms->check,axcb);
	}
}
