/*  State tables for AX25 level 2  */

#include "ax25.h"
#include "ax25l2.h"

/*  Tables for command frames received */

struct change i_cmdp[] = {			/* I frame with poll */
	0,		DM,	0,
	0,		255,	0,
	0,		FRMR,	0,
	1,		DM,	0,
	0,		RR,	0,
	5,		RR,	0,
	0,		RR,	0,
	0,		RNR,	0,
	0,		RR,	0,
	0,		RNR,	0,
	0,		RNR,	0,
	0,		RR,	0,
	0,		RNR,	0,
	0,		RNR,	0,
	9,		RR,	0,
	0,		RNR,	0
};
struct change i_cmd[] = {			/* I frame without poll */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		I,	CMD+RESPOND,
	5,		I,	CMD+RESPOND,
	0,		I,	CMD+RESPOND,
	0,		RNR,	0,
	0,		RR,	0,
	0,		RNR,	0,
	0,		RNR,	0,
	0,		RR,	0,
	0,		RNR,	0,
	0,		RNR,	0,
	9,		RR,	0,
	0,		RNR,	0
};
struct change rr_cmdp[] = {		/* RR command with poll */
	0,		DM,	0,
	0,		255,	0,
	0,		FRMR,	0,
	1,		DM,	0,
	0,		RR,	0,
	0,		RR,	0,
	0,		RR,	0,
	0,		RNR,	0,
	5,		RR,	0,
	8,		RNR,	0,
	0,		RNR,	0,
	7,		RR,	0,
	11,		RNR,	0,
	0,		RNR,	0,
	6,		RR,	0,
	14,		RNR,	0
};
struct change rr_cmd[] = {			/* RR command without poll */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		I,	0,
	0,		I,	0,
	0,		I,	0,
	0,		I,	0,
	5,		I,	0,
	8,		I,	0,
	0,		I,	0,
	7,		I,	0,
	11,		I,	0,
	0,		I,	0,
	6,		I,	0,
	14,		I,	0
};
struct change rej_cmdp[] = {		/* REJ command with poll */
	0,		DM,	0,
	0,		255,	0,
	0,		FRMR,	0,
	1,		DM,	0,
	0,		RR,	0,
	0,		RR,	0,
	0,		RR,	0,
	0,		RNR,	0,
	5,		RR,	0,
	8,		RNR,	0,
	0,		RNR,	0,
	7,		RR,	0,
	11,		RNR,	0,
	0,		RNR,	0,
	6,		RR,	0,
	14,		RNR,	0
};
struct change rej_cmd[] = {		/* REJ command without poll */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		I,	CMD,
	0,		I,	CMD,
	0,		I,	CMD,
	0,		I,	CMD,
	5,		I,	CMD,
	8,		I,	CMD,
	0,		I,	CMD,
	7,		I,	CMD,
	11,		I,	CMD,
	0,		I,	CMD,
	6,		I,	CMD,
	14,		I,	CMD
};
struct change rnr_cmdp[] = {		/* RNR command with poll */
	0,		DM,	0,
	0,		255,	0,
	0,		FRMR,	0,
	1,		DM,	0,
	9,		RR,	0,
	15,		RR,	0,
	12,		RR,	0,
	10,		RNR,	0,
	0,		RR,	0,
	0,		RNR,	0,
	13,		RNR,	0,
	0,		RR,	0,
	0,		RNR,	0,
	16,		RNR,	0,
	0,		RR,	0,
	0,		RNR,	0
};
struct change rnr_cmd[] = {		/* RNR command without poll */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	9,		255,	0,
	15,		255,	0,
	12,		255,	0,
	10,		255,	0,
	0,		255,	0,
	0,		255,	0,
	13,		255,	0,
	0,		255,	0,
	0,		255,	0,
	16,		255,	0,
	0,		255,	0,
	0,		255,	0
};
struct change sabm_cmd[] = {		/* SABM frame with or without poll */
	5,		UA,	RSTT1,
	5,		UA,	RSTT1,
	5,		UA,	RSTT1,
	1,		DM,	RSTT1,
	0,		UA,	RSTT1,
	5,		UA,	RSTT1,
	5,		UA,	RSTT1,
	0,		UA,	RSTT1,
	5,		UA,	RSTT1,
	8,		UA,	RSTT1,
	8,		UA,	RSTT1,
	5,		UA,	RSTT1,
	8,		UA,	RSTT1,
	8,		UA,	RSTT1,
	5,		UA,	RSTT1,
	5,		UA,	RSTT1
};
struct change disc_cmd[] = {		/* DISC frame with or without poll */
	0,		DM,	RSTT1,
	1,		DM,	RSTT1,
	1,		UA,	RSTT1,
	1,		UA,	RSTT1,
	1,		UA,	RSTT1,
	1,		UA,	RSTT1,
	1,		UA,	RSTT1,
	1,		UA,	RSTT1,
	1,		UA,	RSTT1,
	1,		UA,	RSTT1,
	1,		UA,	RSTT1,
	1,		UA,	RSTT1,
	1,		UA,	RSTT1,
	1,		UA,	RSTT1,
	1,		UA,	RSTT1,
	1,		UA,	RSTT1
};

/* Tables for response frames received */

struct change rr_respf[] = {		/* RR response with final */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		I,	CMD,
	0,		I,	CMD,
	5,		I,	CMD+RSTT1,
	0,		I,	CMD,
	5,		I,	CMD,
	8,		I,	CMD,
	8,		I,	CMD+RSTT1,
	5,		I,	CMD+RSTT1,
	8,		I,	CMD+RSTT1,
	0,		I,	CMD,
	6,		I,	CMD,
	14,		I,	CMD
};
struct change rr_resp[] = {		/* RR response without final */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		I,	CMD,
	0,		I,	CMD,
	0,		I,	CMD,
	0,		I,	CMD,
	5,		I,	CMD,
	8,		I,	CMD,
	0,		I,	CMD,
	7,		I,	CMD,
	11,		I,	CMD,
	0,		I,	CMD,
	6,		I,	CMD,
	14,		I,	CMD
};
struct change rej_respf[] = {		/* REJ response with final */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		I,	CMD,
	0,		I,	CMD,
	5,		I,	CMD+RSTT1,
	0,		I,	CMD,
	5,		I,	CMD,
	8,		I,	CMD,
	8,		I,	CMD+RSTT1,
	5,		I,	CMD+RSTT1,
	8,		I,	CMD+RSTT1,
	0,		I,	CMD,
	6,		I,	CMD,
	14,		I,	CMD
};
struct change rej_resp[] = {		/* REJ response without final */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		I,	CMD,
	0,		I,	CMD,
	0,		I,	CMD,
	0,		I,	CMD,
	5,		I,	CMD,
	8,		I,	CMD,
	0,		I,	CMD,
	7,		I,	CMD,
	11,		I,	CMD,
	0,		I,	CMD,
	6,		I,	CMD,
	14,		I,	CMD
};
struct change rnr_respf[] = {		/* RNR response with final */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	9,		255,	RSTT1+SETT1,
	15,		255,	RSTT1+SETT1,
	9,		255,	RSTT1+SETT1,
	10,		255,	RSTT1+SETT1,
	0,		255,	RSTT1+SETT1,
	0,		255,	RSTT1+SETT1,
	10,		255,	RSTT1+SETT1,
	0,		255,	RSTT1+SETT1,
	0,		255,	RSTT1+SETT1,
	16,		255,	RSTT1+SETT1,
	0,		255,	RSTT1+SETT1,
	0,		255,	RSTT1+SETT1
};
struct change rnr_resp[] = {		/* RNR response without final */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	9,		255,	SETT1,
	15,		255,	SETT1,
	12,		255,	0,
	10,		255,	SETT1,
	0,		255,	SETT1,
	0,		255,	SETT1,
	13,		255,	0,
	0,		255,	0,
	0,		255,	0,
	16,		255,	SETT1,
	0,		255,	SETT1,
	0,		255,	SETT1
};
struct change ua_resp[] = {		/* UA response with or without final */
	0,		255,	0,
	5,		255,	RSTT1,
	0,		255,	0,
	1,		255,	RSTT1,
	0,		255,	0,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1
};
struct change dm_resp[] = {		/* DM frame with or without final */
	0,		255,	0,
	1,		255,	RSTT1,
	0,		255,	0,
	1,		255,	RSTT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1
};
struct change frmr_resp[] = {		/* FRMR frame with or without final */
	0,		255,	0,
	0,		255,	0,
	2,		SABM,	CMD+RSTT1+SETT1,
	1,		255,	RSTT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1,
	2,		SABM,	CMD+RSTT1+SETT1
};

/* Tables for asynchronous events */

struct change go_busy[] = {		/* DXE goes busy */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	8,		RNR,	0,
	14,		RNR,	0,
	11,		RNR,	0,
	0,		255,	0,
	10,		RNR,	0,
	0,		255,	0,
	0,		255,	0,
	13,		RNR,	0,
	0,		255,	0,
	0,		255,	0,
	16,		RNR,	0,
	0,		255,	0
};
struct change clr_busy[] = {		/* DXE clears busy */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	5,		RR,	0,
	0,		255,	0,
	9,		RR,	0,
	7,		RR,	0,
	0,		255,	0,
	12,		RR,	0,
	6,		RR,	0,
	0,		255,	0,
	15,		RR,	0
};
struct change t1_exp[] = {			/* T1 timer expires */
	0,		255,	0,
	0,		SABM,	CMD+PF+SETT1,
	0,		FRMR,	0,
	0,		DISC,	CMD+PF+SETT1,
	7,		RR,	CMD+PF+SETT1,
	7,		RR,	CMD+PF+SETT1,
	0,		RR,	CMD+PF+SETT1,
	11,		RNR,	CMD+PF+SETT1,
	12,		RR,	CMD+PF+SETT1,
	13,		RNR,	CMD+PF+SETT1,
	0,		RNR,	CMD+PF+SETT1,
	0,		RR,	CMD+PF+SETT1,
	0,		RNR,	CMD+PF+SETT1,
	11,		RNR,	CMD+PF+SETT1,
	12,		RR,	CMD+PF+SETT1,
	11,		RNR,	CMD+PF+SETT1
};
struct change t1_expv1[] = {		/* T1 timer expires for version 1 protocol */
	0,		255,	0,
	0,		SABM,	PF+SETT1,
	0,		FRMR,	0,
	0,		DISC,	PF+SETT1,
	7,		I,	PF+SETT1,
	7,		I,	PF+SETT1,
	0,		I,	PF+SETT1,
	11,		I,	PF+SETT1,
	12,		I,	PF+SETT1,
	13,		I,	PF+SETT1,
	0,		I,	PF+SETT1,
	0,		I,	PF+SETT1,
	0,		I,	PF+SETT1,
	11,		I,	PF+SETT1,
	12,		I,	PF+SETT1,
	11,		I,	PF+SETT1
};
struct change t3_exp[] = {			/* T3 timer expires */
	0,		255,	0,
	0,		SABM,	CMD+PF+SETT1,
	0,		FRMR,	0,
	0,		DISC,	CMD+PF+RSTT1+SETT1,
	7,		RR,	CMD+PF+RSTT1+SETT1,
	7,		RR,	CMD+PF+RSTT1+SETT1,
	0,		255,	0,
	11,		RNR,	CMD+PF+RSTT1+SETT1,
	12,		RR,	CMD+PF+RSTT1+SETT1,
	13,		RNR,	CMD+PF+RSTT1+SETT1,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	11,		RNR,	CMD+PF+RSTT1+SETT1,
	12,		RR,	CMD+PF+RSTT1+SETT1,
	11,		RNR,	CMD+PF+RSTT1+SETT1
};
struct change inv_ns[] = {			/* Invalid send-sequence number received */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	6,		REJ,	0,
	0,		255,	0,
	0,		255,	0,
	0,		RNR,	0,
	15,		REJ,	0,
	0,		RNR,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		RNR,	0,
	0,		255,	0,
	0,		255,	0
};
struct change inv_nsv1[] = {			/* Invalid send-sequence number received */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	6,		REJ,	0,
	0,		REJ,	0,
	0,		REJ,	0,
	0,		RNR,	0,
	15,		REJ,	0,
	0,		RNR,	0,
	0,		RNR,	0,
	0,		REJ,	0,
	0,		RNR,	0,
	0,		RNR,	0,
	0,		REJ,	0,
	0,		RNR,	0
};
struct change n2_exceeded[] = {	/* N2 count exceeded */
	0,		255,	0,
	1,		255,	0,
	4,		DISC,	CMD+PF+RSTT1+SETT1,	/* was 2,sabm */
	1,		255,	0,
	0,		255,	0,
	4,		DISC,	CMD+PF+RSTT1+SETT1,
	4,		DISC,	CMD+PF+RSTT1+SETT1,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	4,		DISC,	CMD+PF+RSTT1+SETT1,
	4,		DISC,	CMD+PF+RSTT1+SETT1,
	4,		DISC,	CMD+PF+RSTT1+SETT1,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0
};
struct change val_ns[] = {			/* Valid N(S) received */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	RSTT1,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0
};
struct change val_nsv1[] = {			/* Valid N(S) received, version 1 */
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	0,		255,	RSTT1,
	0,		255,	RSTT1,
	5,		255,	RSTT1,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0,
	8,		255,	RSTT1,
	9,		255,	RSTT1,
	10,		255,	RSTT1,
	0,		255,	0,
	0,		255,	0,
	0,		255,	0
};
