/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */
#include "data.h"
#include "buffer.h"
#include "iface.h"
#include "timer.h"
#include "ax25.h"
#include "ax25l2.h"
#include "l3struc.h"
#include "x25cause.h"
#include "l3calls.h"
#include "tx.h"
#include "config.h"

#define L3SIZ 32
extern struct ax25_addr	L3CALL[L3SIZ],
			L3DIGI[L3SIZ];
extern unsigned char	L3PORT[L3SIZ],
			L3_WI[L3SIZ],
			L3_MAXVC[L3SIZ];
extern int		L3DELAY[15];

extern unsigned char node_addr[];
extern struct ax25_parms *l2parms, *l3parms;
extern struct datastr *l2_user_info, l2_info;

struct dnic {
	unsigned char adr[3];	/* DNIC in X.121 format */
	unsigned char *rt;	/* Routing table root */
	int len;		/* Length of routing data */
};

extern unsigned char NPAroute[];
extern unsigned char ndnics;
extern struct dnic DNics[];

extern unsigned char L1BSS, L1BSSL, Init_bss;

struct datastr *mkpkt();
int check_x121();
unsigned char *malloc(), *calloc();
void free();

int bss_i(), usr_info(), rt_init(), examod_i();
int buf();

#define READONLY 0x80
#define MUSTCLEAR 0x40
#define RESET 0x20
#define DYNAMIC 0x10
#define FCNCALL 0x01

  /* MUSTCLEAR says to update this we need to clear all links */
  /* RESET says we can reset this after a read */
  /* DYNAMIC says the Object is a pointer pointing to malloc'ed memory */
  /* FCNCALL says to call *fcn() to copy byte */

struct inftab {
	void *obj;		/* Pointer to the object we are working with */
	int len;		/* It's length (number of items) */
	int size;		/* it's size (bytes/item) */
	char flags;		/* Special handling flags */
	int (*init)();		/* Set up vars for get/put */
	int (*fcn)();		/* Function to handle one byte */
};
struct inftab inf[MAXOBJ] = {
		{&l2_user_info,1,0,DYNAMIC+FCNCALL,usr_info,buf},
		{L3CALL,L3SIZ,AXALEN,0,bss_i,NULL},
		{L3DIGI,L3SIZ,AXALEN,0,bss_i,NULL},
		{L3PORT,L3SIZ,1,0,bss_i,NULL},
		{L3_MAXVC,L3SIZ,1,0,bss_i,NULL},
		{L3_WI,L3SIZ,1,0,bss_i,NULL},
		{L3DELAY,16,2,0,bss_i,NULL},
		{&Init_bss,1,54,READONLY,bss_i,NULL},
		{node_addr,1,7,0,bss_i,NULL},
		{&l2parms,1,sizeof(struct ax25_parms),DYNAMIC,bss_i,NULL},
		{&l3parms,1,sizeof(struct ax25_parms),DYNAMIC,bss_i,NULL},
		{NPAroute,1,120,0,bss_i,NULL},
		{DNics,5,5,0,bss_i,NULL},
		{DNics,0,5,DYNAMIC,rt_init,NULL},
		{&L1BSS,1,(int)&L1BSSL,READONLY,bss_i,NULL},
		{NULL,0,1,DYNAMIC,examod_i,NULL}
	};

#define WRITE 01
#define READ 02
#define READR 03
#define MAXCMD 03

#if 0
char *errmsg[]={
	"Invalid command\n",
	"Invalid Object specified\n",
	"No working memory!\n",
	"Bad checksum!\n",
	"Not implemented yet, sorry\n"
	"Odd number of data bytes for type\n"
	"Item is Read-Only"
};
#endif

char config_busy;
char config_state;
struct datastr *config_in;
char config_cmd;
char config_obj;
int config_len;
int config_pos;
unsigned char *config_data, *save_config_data;
unsigned char config_checksum;
char config_high;
unsigned char config_chv;

int
hexch(ch)
unsigned char ch;
{
	if (ch>0x39) ch -= 7;
	if (ch<0x30 || ch>0x3f) return -1;
	return (ch-0x30);
}

putxch(p,val)
register struct datastr *p;
int val;
{
	static char v;

	v= (val >> 4) &0x0f;
	if (v > 9) v += 7;
	bappch(p,v+0x30);
	v=(val & 0x0f);
	if (v > 9) v += 7;
	bappch(p,v+0x30);
}

putxint(p,val)
register struct datastr *p;
int val;
{
	putxch(p,val);		/* Low byte Low address */
	putxch(p,(val>>8));
}

int
con_config(vc)
register struct VCS *vc;
{
	static int (*fn)();

	vc=vc->peer;
	if (!vc) return;
	fn=vc->SEND;
	puthex2("con",vc,"at",fn);
	(*fn)(vc,dup_pkt(&l2_info));
	if (config_busy) {
		(*fn)(vc,mkpkt("Busy\r"));
		set_p(vc,P6,Number_Busy);
	}
	else {
		config_busy=TRUE;
		config_state=1;
	}
}

int
clr_config(vc,c)
struct VCS *vc;
int c;
{
	config_busy=FALSE;
	set_p(vc,P1,0);
}

int
recv_config(vc,pkt)
register struct VCS *vc;
struct datastr *pkt;
{
	if (vc && (vc=vc->peer)) (*vc->SEND)(vc,NULL);	/* Kick transmitter */
}

int
send_config(vc, pkt)
register struct VCS *vc;
struct datastr *pkt;
{
	static int i, ch;

	puthex2("send fig",vc,"",pkt);
	vc_queue_data(vc,pkt);
	while (config_in = vc->tx_queue) /* There is stuff waiting */ {
		vc->tx_queue=config_in->next;
		config_in->next=NULL;
		while ((ch=bgetch(&config_in)) != EOF) {
			puthex2("loop",ch,"",config_in);
			if (ch > 0x20) /* Allow Control chars to pass */ {
			i=hexch(ch);
			if (ch == 0x3a || i<0 ) config_state=1;
			if (config_state > 1) {
				if (config_high) config_chv = i<<4;
				else config_chv += i;
				config_high = !config_high;
			} else config_chv=ch;
			if (config_state == 1 || config_high)
				config_next(vc, config_chv);
			}
		}
	}
	return 256;	/* We are never busy! */
}

/* config_state has the following meaning 
/*	1 - Searching for a : (Start of record)
/*	2 - Next byte read is the command
/*	3 - Next byte read is the object we are working with
/*	4 - Next byte read is the low byte of length !Low byte Low address!
/*	5 - Next byte read is the high byte of length
/*	6 - Next byte read is additional data
/*	7 - Processing request
*/

config_next(vc,chv)
struct VCS *vc;
unsigned char chv;
{
	static char err;
	static int len;
	register struct VCS *vcp;
	static int (*fcn)();
	static struct datastr *p;
	static struct inftab *rec;
	int do_config();

	err=0;
	vcp=vc->peer;
	fcn=vcp->SEND;
	puthex2("next",config_state,"ch",chv);
	switch (config_state) {
		case 1: if (chv != 0x3a) config_state=0;
			config_high = TRUE;
			break;
		case 2: config_cmd=chv;
			if (chv>MAXCMD) err=1;
			break;
		case 3: config_obj=chv;
			if (chv>MAXOBJ) err=2;
			break;
		case 4: config_len= chv;	/* Low byte Low address */
			break;
		case 5: config_len += (chv<<8);
			config_pos=0;
			config_checksum=0;
			config_data=NULL;
			rec = &inf[config_obj];
			len= rec->size * rec->len;
			if (rec->flags & DYNAMIC) len=config_len;
			if (len) /* && config_cmd == WRITE) */ {
				config_data=save_config_data=calloc(1,len);
				if (config_data == NULL) return 3;
			}
			if (config_len == 0) config_state++; /* No data field */
			break;
		case 6: config_data[config_pos]=chv;
			config_pos++;
			config_checksum += chv;
			if (config_pos != config_len) config_state--;
			break;
		case 7: config_checksum += chv;
			if (config_checksum != 0) err=4;
			else err=do_config(vcp);
			if (!err) (*fcn)(vcp,mkpkt("OK\r"));
			if (save_config_data) free(save_config_data);
			config_state=0;
			break;
	}
	if (err) {
		if (vcp) {
			p=mkpkt("Error");
			putxch(p,err);
			bappch(p,0x0d);
			(*fcn)(vcp,p);
		}
		config_state=0;
	}
	config_state++;
	puthex2("exit",config_state,"err",err);
}

int
do_config(vc)
struct VCS *vc;
{
	static int i, j, len;
	static struct datastr *bp;
	register struct inftab *rec;
	static unsigned char *ch, *ch1;

	puthex2("do it",config_cmd,"",config_obj);
	rec = &inf[config_obj];
	len = (*rec->init)(rec, &ch);
	if (len < 0) return -len;
	if (config_cmd == WRITE) {
		if (rec->flags & READONLY) return 7;
		ch1 = (unsigned char *)config_data;
		while (len>0) {
			if (rec->flags & FCNCALL) {
				(*rec->fcn) (rec, &ch, &ch1);
				ch1++;
			}
			else *ch++ = *ch1++;
			len--;
		}
		sumchk();	/* Reset the checksum for bbRAM */
	}
	else if (config_cmd == READ) {
		bp=new_buffer((len*2)+10);
		if (!bp) return 3;
		bappch(bp,0x3a);
		putxch(bp,config_obj);	/* Identify what we are sending */
		putxint(bp,len);		/* and how big it is */
		config_checksum=0;
		for (i=0;i < len ;i++) {
			if (rec->flags & FCNCALL)
				j = (*rec->fcn)(rec, &ch, NULL);
			else j = *ch++;
			config_checksum += j;
			putxch(bp,j);
		}
		putxch(bp,-config_checksum);
		bappch(bp,0x0d);
		(*vc->SEND)(vc,bp);
	}
	else return 5;
	return 0;
}

usr_info(rec,ch)
struct inftab *rec;
struct datastr **ch;
{
	register struct datastr *bp;
	static int len;

	if (config_cmd == WRITE) {
		if (l2_user_info != &l2_info) free_pkt(l2_user_info);
		l2_user_info=dup_pkt(&l2_info);
		len = config_len;
		bp=new_buffer(len);
		if (!bp) return -1;
		l2_user_info=binsert(l2_user_info,bp);
	}
	else /* It's Read (or ReadR) */ {
		bp=dup_pkt(l2_user_info);	/* Leave original alone */
		if (!bp) len = -1;
		else len = buflen(bp);
	}
	*ch = bp;
	return len;
}

int
buf(rec,bpx,ch)
struct inftab *rec;
struct datastr **bpx;
unsigned char **ch;
{
	if (config_cmd == WRITE) bappch(*bpx,**ch);
	else /* Read */ {
		if (*bpx == NULL) return EOF;
		return (bgetch(bpx));
	}
}

int
bss_i(rec,ch)
struct inftab *rec;
unsigned char **ch;
{
	*ch= (unsigned char *)rec->obj;
	if (rec->flags & DYNAMIC) *ch= *(unsigned char **)ch;
	return (rec->size * rec->len);
}


int
rt_init(rec,ch)
struct inftab *rec;
unsigned char **ch;
{
	static int i, j, k, l, nd, c;
	static struct dnic dn;

	for (i=0;i<3;i++) {
		if (config_len <= 0) return -6;
		config_len--;
		c = *config_data++;
		dn.adr[i] = c;
	}
	for (i=0;i<ndnics;i++) {
		if (check_x121(DNics[i].adr,dn.adr) >= 0) break;
	}
	if (config_cmd == WRITE) {
		*ch = NULL;	/* Invalid DNIC */
		if (i<ndnics) free_stuff(&DNics[i].rt);
		l = config_len;
		if (config_len == 0) /* Delete DNIC */ {
			if (i == ndnics) return 0; /* Deleted Invalid DNIC */
			nd = ndnics-1;
			if (i == nd) ndnics--;
			else /* Have to copy rest... */ {
				for (j=i;j<nd;j++) {
					for (k=0;k<3;k++)
					   DNics[j].adr[k] = DNics[j+1].adr[k];
					DNics[j].rt = DNics[j+1].rt;
				}
			}
		}
		else /* Add Dnic */ {
			if (i == ndnics) {
				ndnics++;
				for (j=0;j<3;j++) DNics[i].adr[j] = dn.adr[j];
			}
			else free_stuff(&DNics[i].rt);
			DNics[i].rt = *ch = malloc(config_len);
			DNics[i].len = config_len;
		}
	}
	else /* READ */ {
		if (i == ndnics) return -2;	/* Invalid DNIC */
		*ch = DNics[i].rt;
		l = DNics[i].len;
	}
	return l;
}

int
examod_i(rec,ch)
struct inftab *rec;
unsigned char **ch;
{
	static int len;
	register unsigned int *dta;

	if (config_len < 4) return -6;
	dta = (unsigned int *) config_data;
	*ch = (unsigned char *) *dta;
	len = *(dta+1);
	config_data += 4;
	config_len -= 4;
	return len;
}
