/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */
/* This file contains the routines that parse x.25 facilities field */


fac_len(ptr)
register unsigned char ptr[];
{
	static unsigned char l;

	if (! ptr[0]) /* Facility marker?? error */ l=255;
	else {
		l=( ptr[0] & 0xc0) >> 6;
		if (l==3) /* Variable length */
			l= ptr[1] + 2;
		else l += 2;
	}
	return l;
}

x25_facilities(vc, ptrx)
struct VCS *vc;
unsigned char ptrx[];
{
	static unsigned char len, fl ,marker;
	static int i;
	static unsigned char *ptr;

	ptr=ptrx;
	marker=1; /* X.25 Facilities, 0=National, 15=CCITT */
	i=1;
	len= ptr[0];
	while (i<len) /* More stuff to copy */ {
		if (ptr[i]==0) /* Facility Marker */ {
			i++;
			marker= ptr[i++];
		}
		else /* Which set are we in now? */ {
			fl=(unsigned char)fac_len(&ptr[i]);
			if (i+fl-1>len) break; /* error */
			if (marker==0) national_fac(vc,&ptr[i]);
			else if (marker==15) ccitt_fac(vc,&ptr[i]);
			i += fl;
		}
	}
	if (i<len) return -1;
	return len;
}

national_fac(vc,ptr)
register struct VCS *vc;
unsigned char *ptr;
{
	static int i;
	static struct VCS *vcp;

	vcp=vc->peer;
	switch ((*ptr++) & 0x0ff) {
	case 0x7f: /* Routing tag, to avoid loops */
		i=(*ptr++);
		vc->rand = vcp->rand = (i<<8) + (*ptr++);
		break;
	case 0x0e9: /* Destination Digipeater */
		fac_call(ptr,&vcp->digi);
		break;
	case 0x0eb: /* Source Digipeater */
		fac_call(ptr,&vc->digi);
		break;
	}
}

ccitt_fac(vc,ptr)
register struct VCS *vc;
unsigned char *ptr;
{
	static int i;
	static unsigned char *ch;
	ch=NULL;
	switch ((*ptr++) & 0x0ff) {
	case 0x0c9: /* Destination NSAP/Callsign */
		ch=(unsigned char *)vc->peer->nsap;
		break;
	case 0x0cb: /* Source NSAP/Callsign */
		ch=(unsigned char *)vc->nsap;
		break;
	}
	if (ch) /* Valid Facility code */ {
		i= *ptr++;	/* Facility Length */
		while (i>=0) {
			*ch++ = *ptr++;
			i--;
		}
	}
}

fac_call(ptr,str)
register unsigned char *ptr;
unsigned char str[];
{
	static int i;

	i= (*ptr++) & 0x3f;
	if (i == AXALEN) /* Is length correct? */
		for (i=0; i<AXALEN; i++)
			str[i]= *ptr++;
}

put_fac(vcx,pkt)
struct VCS *vcx;
struct datastr *pkt;
{
	static unsigned char adrstr[32];
	static unsigned char i, j, k, l;
	register unsigned char *chp;
	static struct VCS *vc, *vcp;

	vc=vcx;
	vcp=vc->peer;
	adrstr[0]=0;
	adr_copy(0,vc->addr,0,adrstr);
	adr_copy(0,vcp->addr,1,adrstr);
	i=(adrstr[0] & 0x0f) + ((adrstr[0] & 0x0f0) >>4);
	i= ( (i + (i & 1)) >> 1 ) + 1;
	for (j=0;j<i;j++)
		bappch(pkt,adrstr[j]);

	chp = str_work;

	*chp++ = 0;
	*chp++ = 0x0f; /* CCITT Marker */
	l=(vc->nsap[0]+1)>>1;	/* Destination NSAP */ 
	*chp++ =0xc9;
	*chp++ =l+1;
	for (k=0;k<=l;k++)	/* ISO address */
		*chp++ =vc->nsap[k];

	l=(vcp->nsap[0]+1)>>1;	/* Source NSAP */
	*chp++ =0xcb;
	*chp++ =l+1;
	for (k=0;k<=l;k++)
		*chp++ =vcp->nsap[k];

	if (vc->digi.call[0] || vcp->digi.call[0] || vc->rand) {
		*chp++ =0;
		*chp++ =0; /* National Marker */
		if (vc->digi.call[0] != 0) /* Dest Digi */ {
			*chp++ =0xe9;
			*chp++ =AXALEN;
			for (k=0;k<ALEN;k++)
				*chp++ =vc->digi.call[k];
			*chp++ =vc->digi.ssid;
		}
		if (vcp->digi.call[0] != 0) /* Source Digi */ {
			*chp++ =0xeb;
			*chp++ =AXALEN;
			for (k=0;k<ALEN;k++)
				*chp++ =vcp->digi.call[k];
			*chp++ =vcp->digi.ssid;
		}
		*chp++ =0x7f;
		*chp++ =(vc->rand >> 8);
		*chp++ =vc->rand &0xff;
	}

	if ((l=vc->data[0]) != 0) /* There is call user data */
		for (k=1;k<=l;k++)
			*chp++ =vc->data[k];
	l= chp - str_work;
	bappch(pkt,l);
	for (k=0;k<l;k++) bappch(pkt,str_work[k]);
}
/* I should have checked the appends to see if it ever failed */

get_nsap(vc,stnx)
struct VCS *vc;
struct ax25_addr *stnx;
{
	static int i,j,max;
	static struct ax25_addr *stn;
	static unsigned char *adr;

	/* Should check for ICD 0011, the RATS ICD!!! */

	stn=stnx;
	adr=vc->nsap;
	max= (*adr + 1)>>1;
	j = 4+1 + (adr[4] + 1 >> 1) ; /* Source pos */
	if (j>=max) return -1;
	for (i=0;i<ALEN;i++) {
		if ((j==max) || (adr[j] == 0x2d)) stn->call[i]=0x40;
		else stn->call[i]=(adr[j++]<<1);  /* Normal ch */
	}
	stn->ssid=0;
	if (adr[j++] == 0x2d && j<=max && adr[j]>=0x30 && adr[j]<=0x39) {
		stn->ssid = (adr[j++]-0x30)<<1;
		if (j<=max && adr[j]>=0x30 && adr[j]<=0x39) {
			stn->ssid = (stn->ssid*10) + ((adr[j] - 0x30)<<1);
		}
		if (stn->ssid > 30) stn->ssid=0;
	}
	return 0;
}

mk_nsap(vc,stnx)
struct VCS *vc;
struct ax25_addr *stnx;
{
	static struct ax25_addr *stn;
	static unsigned char *addr, i, j, ch, len;
	register unsigned char *nsapadr;

	stn=stnx;
	nsapadr=&vc->nsap[1];
	addr=&vc->addr[0];
	j= (*addr + 1)>>1;
	*nsapadr++ = 0x47;
	*nsapadr++ = 0x00;
	*nsapadr++ = 0x11;
	for (i=0;i<=j;i++) *nsapadr++ = *addr++;
	len = 0;
	for (i=0;i<ALEN;i++) {
		ch=stn->call[i];
		if (ch != 0x40) {
			*nsapadr++ = (ch>>1);
			len++;
		}
	}
	*nsapadr++ = 0x2d;	/* "-" */
	len++;
	j=(stn->ssid & SSID)>>1;
	if (j>9) {
		*nsapadr++ = 0x31;
		len++;
		j -= 10;
	}
	*nsapadr++ = (0x30 + j);
	len++;
	vc->nsap[0]=(len<<1) + 8 + (vc->addr[0] + 1) & ~1;
}
