*TITLE	HDWINT - 'Hardware init and linking module'

;  HDWINT performs the necessary system initialization.
;  The SIO interrupt service routine entry points are
;  also here so that the SIO data structure for the
;  interrupting channel can be external to the SIO
;  driver (HDLSIO).  This allows that module to be
;  generic.

*INCLUDE RCONFIG.LIB
*INCLUDE ENVIR.DEF
*INCLUDE STRUCT.LIB

MSYNC	EQU	10H
MCTS	EQU	20H

;  External routine declaration

	GLOBAL	HDWINT, IPAGE
	GLOBAL	HDLI,XSKED ;,CDS		;HDLSIO callable subs
	GLOBAL	I_TBE,I_RBF,I_EXT,I_SRX	;HDLSIO interrupt subs
	COND	TNC2 .eq. FALSE
	GLOBAL	RATES,TXDLYS
	ENDC
	GLOBAL	TICK, SPRINTC, CLKBIT, TICKCNT

	PSECT	bss
SPRINTC: DEFS 1
	PSECT	text

*HEADING 'Initialization Subroutine'

;  Initialize the interrupt vector page

HDWINT:
	LD	HL,IPAGE	;Base of vectors
	LD	DE,SPRINT	;Spurious handler
	LD	B,128		;Number of vectors
ILOOP:	LD	(HL),E		;Set vector to SPRINT
	INC	HL
	LD	(HL),D
	INC	HL
	DJNZ	ILOOP

;  Initialize the SIO vector register  (Now done during Channel Init.)

;	LD	A,2		;Set the SIO vector reg
;	OUT	(SIOBC),A
;	LD	A,0
;	OUT	(SIOBC),A

	COND	TNC2 .eq. FALSE
;  Initialize the timer CTC

	LD	A,3		;Reset the CTC channels
	OUT	(CTC),A
	OUT	(CTC+1),A
	OUT	(CTC+2),A
	OUT	(CTC+3),A
	LD	A,10H		;Set the interrupt vector
	OUT	(CTC),A
	LD	A,CTCMOD	;Set up the timer CTC
	OUT	(CTCCH),A
	LD	A,CTCDIV
	OUT	(CTCCH),A

;  Initialize the baud rate generators

	LD	A,(RATES)	;Get channel 0 speed
	OUT	(0CH),A		;Set SIO B rate
	LD	A,(RATES+2)	;Get channel 1 speed
	OUT	(0),A		;Set SIO A rate
	ENDC

;  Initialize the CPU interrupt logic and set the vectors

	LD	DE,IPAGE	;Set the vectors
	LD	HL,IV
	LD	BC,IVL
	LD	A,D		;Set the I register
	LD	I,A
	LDIR
	IM	2		;Interrupt mode 2

;  Initialize the SIO data structures

	LD	HL,SDSI0	;Copy from ROM
	LD	DE,SDS0		;To RAM
	LD	BC,SDSILEN
	LDIR

	COND	NUMCH .gt. 1
	LD	HL,SDSI1	;Copy from ROM
	LD	DE,SDS1		;To RAM
	LD	BC,SDSILEN
	LDIR
	ENDC

	COND	NUMCH .gt. 2
	LD	HL,SDSI2	;Copy from ROM
	LD	DE,SDS2		;To RAM
	LD	BC,SDSILEN
	LDIR
	ENDC

	COND	NUMCH .gt. 3
	LD	HL,SDSI3	;Copy from ROM
	LD	DE,SDS3		;To RAM
	LD	BC,SDSILEN
	LDIR
	ENDC

	COND	0
;  Fill in the TXDLY values

	LD	HL,SDS0+TXDLY	;Point to first TXDLY
	LD	DE,TXDLYS	;Point to ROM values
	LD	B,NUMCH		;Number of times to loop
TXDLOP:	LD	A,(DE)		;Get low byte of value
	INC	DE
	LD	(HL),A		;Store in data structure
	INC	HL
	LD	A,(DE)		;Get high byte
	INC	DE
	LD	(HL),A		;Store
	PUSH	DE
	LD	DE,SDSLEN-1	;Point to next data structure
	ADD	HL,DE
	POP	DE
	DJNZ	TXDLOP		;Loop 'til done
	ENDC

	LD	A,MSYNC		;Default Clock bit is TNC  (SYNC)
	LD	(CLKBIT),A	;Save it
	LD	A,N10MS
	LD	(TICKCNT),A	;Initialize Number of Interrupts/Tick

;  What kind of hardware are we using??

        LD      C,8             ;SEE IF IT'S A DR
        LD      A,12
        OUT     (C),A
        LD      A,55H
        OUT     (C),A           ;STORE VALUE IN REG. 12 AND READ IT BACK
        LD      A,12
        OUT     (C),A
        IN      A,(C)
        CP      55H             ;A 55H MEANS IT'S A DR
        JR      NZ,SIOINIT      ;ELSE, NEITHER 220 OR DR, MUST BE A SIO CHIP
	LD	A,09
	LD	C,8
	OUT	(C),A
	LD	A,0C0H		;Master Reset for SCC
	OUT	(C),A
	NOP
	NOP			;Wait for it to happen
	LD	A,MCTS		;DR200 Clock bit (CTS)
	LD	(CLKBIT),A	;Save it
	LD	IY,SDS0		;Port 0
	LD	(IY+DPORT),11	;Data Port - 11
	LD	(IY+CPORT),9	;Control Port - 9
	LD	DE,SCCBI
	LD	(IY+SINIT),E	;Address of Initialization Table
	LD	(IY+SINIT+1),D
	LD	(IY+SINTL),SCCAIL ;Length of Initialization Table
	LD	(IY+FLAG),0	;Not used for Timer
	LD	(IY+WR5REG),0E9H ;SYNC WR5 REG (DTR on CON)
	LD	IY,SDS1		;Port 1
	LD	(IY+DPORT),10	;Data Port - 10
	LD	(IY+CPORT),8	;Control Port - 8
	LD	DE,SCCBI
	LD	(IY+SINIT),E	;Address of Initialization Table
	LD	(IY+SINIT+1),D
	LD	(IY+SINTL),SCCBIL	;Length of Initialization Table
	LD	(IY+FLAG),MTIMER	;Line Used for Timer
	LD	(IY+WR5REG),069H	;NEED SPECIAL WR5 (E9 with DTR off STA)
	LD	A,N10MS/2		;Number of Interrupts/Tick
	LD	(TICKCNT),A

;  Initialize the SIO hardware
SIOINIT:
	LD	IY,SDS0
	CALL	HDLI

	COND	NUMCH .gt. 1
	LD	IY,SDS1
	CALL	HDLI
	ENDC

	COND	NUMCH .gt. 2
	LD	IY,SDS2
	CALL	HDLI
	ENDC

	COND	NUMCH .gt. 3
	LD	IY,SDS3
	CALL	HDLI
	ENDC

	RET

*HEADING 'Interrupt Service Routines'

;  SPRINT - Spurious interrupt service
;  A deadly condition requires a drastic cure.

SPRINT:
	RETI
;	HALT			;Die, you old bastard
;	jp	SPRINT		;and stay here!
;	PUSH	HL
;	LD	HL,SPRINTC	;KEEP A COUNT
;	INC	(HL)
;	JP	NZ,1f
;	DEC	(HL)		;COUNT, BUT DON'T WRAP
;1:	POP	HL

;  SIO interrupt service entries

;XSKED0:
;	LD	IY,SDS0
;	JP	XSKED
;CDS0:
;	LD	IY,SDS0
;	JP	CDS
I_TBE0:
	PUSH	IY
	LD	IY,SDS0
	JR	TBE
I_RBF0:
	PUSH	IY
	LD	IY,SDS0
	JP	RBF
I_EXT0:
	PUSH	IY
	LD	IY,SDS0
	JP	EXT
I_SRX0:
	PUSH	IY
	LD	IY,SDS0
	JP	SRX

	COND	NUMCH .gt. 1
;XSKED1:
;	LD	IY,SDS1
;	JP	XSKED
;CDS1:
;	LD	IY,SDS1
;	JP	CDS
I_TBE1:
	PUSH	IY
	LD	IY,SDS1
	JR	TBE
I_RBF1:
	PUSH	IY
	LD	IY,SDS1
	JP	RBF
I_EXT1:
	PUSH	IY
	LD	IY,SDS1
	JP	EXT
I_SRX1:
	PUSH	IY
	LD	IY,SDS1
	JP	SRX

	ENDC
	COND	NUMCH .gt. 2
;XSKED2:
;	LD	IY,SDS2
;	JP	XSKED
;CDS2:
;	LD	IY,SDS2
;	JP	CDS
I_TBE2:
	PUSH	IY
	LD	IY,SDS2
	JR	TBE
I_RBF2:
	PUSH	IY
	LD	IY,SDS2
	JP	RBF
I_EXT2:
	PUSH	IY
	LD	IY,SDS2
	JP	EXT
I_SRX2:
	PUSH	IY
	LD	IY,SDS2
	JP	SRX

	ENDC
	COND	NUMCH .gt. 3
;XSKED3:
;	LD	IY,SDS3
;	JP	XSKED
;CDS3:
;	LD	IY,SDS3
;	JP	CDS
I_TBE3:
	PUSH	IY
	LD	IY,SDS3
	JR	TBE
I_RBF3:
	PUSH	IY
	LD	IY,SDS3
	JP	RBF
I_EXT3:
	PUSH	IY
	LD	IY,SDS3
	JP	EXT
I_SRX3:
	PUSH	IY
	LD	IY,SDS3
	JP	SRX
	ENDC

TBE:
	EXX
	LD	HL,I_TBE
	JR	SERVICE
RBF:
	EXX
	LD	HL,I_RBF
	JR	SERVICE
EXT:
	EXX
	LD	HL,I_EXT
	JR	SERVICE
SRX:
	EXX
	LD	HL,I_SRX
	JR	SERVICE

SERVICE:
	PUSH	AF		;Hi-Tech Divide routines use AF and AF'
	PUSH	IX
	CALL	VECTOR
	LD	C,(IY+CPORT)	;;
	LD	A,38H		;;
	OUT	(C),A		;;DR200 IGNORES RETI??
	POP	IX
	POP	AF
	EXX
	POP	IY
	EI
	RETI

VECTOR:
	JP	(HL)		;Vector to subroutine

*HEADING 'SIO Data Structures'

	PSECT	bss
	GLOBAL	SDS0
SDS0:	DEFS	SDSLEN
	PSECT	data
SDSI0:	DEFB	0		;Channel number
	DEFB	DPORT0		;SIO data port
	DEFB	CPORT0		;SIO control port
	DEFW	DWAIT0
	DEFW	TXDELAY0
	DEFW	SIOBI		;SYNC CONFIG
	DEFB	SIOBIL		;LEN
	DEFB	0		;Flags
	DEFB	0E9h		;WR5REG (DTR Now On, CON)

	COND	NUMCH .gt. 1
	PSECT	bss
	GLOBAL	SDS1
SDS1:	DEFS	SDSLEN
	PSECT	data
SDSI1:	DEFB	1		;Channel number
	DEFB	DPORT1		;SIO data port
	DEFB	CPORT1		;SIO control port
	DEFW	DWAIT1
	DEFW	TXDELAY1
	DEFW	SIOAI		;ASYNC CONFIG
	DEFB	SIOAIL		;LEN
	DEFB	MFULL+MASYNC+MTIMER+MAFT	;ON A TNC2
	DEFB	06ah		;WR5REG (DTR is Now Off, STA)
	ENDC

	COND	NUMCH .gt. 2
	PSECT	bss
	GLOBAL	SDS2
SDS2:	DEFS	SDSLEN
	PSECT	data
SDSI2:	DEFB	2		;Channel number
	DEFB	DPORT2		;SIO data port
	DEFB	CPORT2		;SIO control port
	DEFW	DWAIT2
	DEFW	TXDELAY2
	DEFW	SIOBI		;SYNC CONFIG
	DEFB	SIOBIL		;LEN
	DEFB	0		;Flags
	DEFB	0		;WR5REG
	ENDC

	COND	NUMCH .gt. 3
	PSECT	bss
	GLOBAL	SDS3
SDS3:	DEFS	SDSLEN
	PSECT	data
SDSI3:	DEFB	3		;Channel number
	DEFB	DPORT3		;SIO data port
	DEFB	CPORT3		;SIO control port
	DEFW	DWAIT3
	DEFW	TXDELAY3
	DEFW	SIOBI		;SYNC CONFIG
	DEFB	SIOBIL		;LEN
	DEFB	0		;Flags
	DEFB	0		;WR5REG
	ENDC


;  Fixed data

;  SIO ASYNC initialization block.

SIOAI:	DEFB	18H		;Reset SIO
	DEFB	2,0		;Set Up Interrupt Vector
	DEFB	14h		;WR 4 + Reset Ext Intr
	DEFB	44H		;16xClk + 1Stop
	DEFB	3		;WR 3
	DEFB	0C1H		;8bit + RxEnable
	DEFB	5		;WR 5
	DEFB	06AH		;DTR Off + 8bit + TxEnable + RTS
;	DEFB	28H		;Reset Tx Intr
;	DEFB	10H		;Reset Ext Intr
;	DEFB	10H		;Reset Ext Intr
;	DEFB	0C0H		;Reset TxUnderun + EOM
	DEFB	11h		;WR 1 + Reset Ext intr
	DEFB	00011111B 	;Enable rx, ext ints
SIOAIL	EQU	$-SIOAI

;  SIO SYNC initialization block.

SIOBI:	DEFB	18H		;Reset SIO
	DEFB	3		;WR 3
	DEFB	11111001B 	;8 bits, enables
	DEFB	14h		;WR 4 + Reset Ext Intr
	DEFB	00100000B 	;SDLC mode, x1 clock
	DEFB	5		;WR 5
	DEFB	01101001B 	;DTR off, RTS off
	DEFB	7		;WR 7
	DEFB	01111110B 	;SDLC flag byte
	DEFB	11h		;WR 1 + Reset Ext Intr
	DEFB	00011111B 	;Enable rx, ext ints
	DEFB	10000000B	;Reset Tx CRC
SIOBIL	EQU	$-SIOBI

;DR200 8530 VALUES FOR INITIALIZAITON
;
SCCBI:	DEFB	4,0A0H		;X32, SDLC, Sync
	DEFB	10,0A0H		;Init CRC, NRZI
	DEFB	7,7EH		;Sync Definitition (No S!!t)
	DEFB	2,0H		;Interrupt Vector
	DEFB	3,0DAH		;8 bit Rx, Enter Hunt, RXCRC, Sync Load Inh.
	DEFB	5,68H		;8 Bit Tx, Tx Enable
	DEFB	11,66H		;RXC = DPLL O/P, TXC = RTxC, TRxC=I,
;					TRxC Out = BRG Out
	DEFB	12,1EH,13,0	;Time Constant (X000 Baud)
	DEFB	14,82H		;BR Generator source???
        DEFB	5,69H		;8 bit Tx, Tx Enable, TX CRC Enable
	DEFB	3,0DBH		;Same as 3 above, RX Enable
	DEFB	14,23H		;Enter Search Mode, Enable BRG
	DEFB	15,0E8H		;Enable Interrupts on 
	DEFB	10H,10H		;Reset External Interupts, Twice
	DEFB	1,13H		;Interrupt on Receive and Xmit
SCCAIL	EQU	$-SCCBI
	DEFB	9,9		;Master Interrupt Enable, Vector Includes Stat
SCCBIL	EQU     $-SCCBI

;  Interrupt vectors

IV:
	DEFW	I_TBE1,I_EXT1,I_RBF1,I_SRX1
	DEFW	I_TBE0,I_EXT0,I_RBF0,I_SRX0
IVL	EQU	$-IV

	END
