/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */
/* incall.c

	These are the incoming call processing routines.
	This will call call_request (in calsetup.c) to setup
	the peer call */

unsigned char call_num;

incall(rec)
struct table_entry *rec;

{
	register struct VCS *vlc;
	static int i, j, ch;
	static unsigned char adrstr[16];
	static struct VCS *vc, *lcp;
	static struct axcb *lnk;
	static struct ax25_addr vlc_adr, lcp_adr, ax_adr1, ax_adr2;
	extern int spclev();

	vlc=lc;
	if (vlc != NULL) /* Call Collision!! */ {
		if (vlc->P == P5) /* Already in collision!! */ {
			set_p(vlc,P6,DTE_Orig+24);
			return;
		}
		if (link->am_dce) /* Clear my call, and handle his! */ {
			if (vlc->peer != NULL) /* Exists */ {
				vlc->peer->peer=NULL;
				set_p(vlc->peer,P6,Remote_Proc+72);
			}
			set_p(vlc,P1,0);
			vlc=NULL;
		}
		else {
			set_p(vlc,P5,0);
			return; /* Ignore request */
		}
	}
	vlc=new_vc();
	if (vlc == NULL) /* No memory!! */ {
		keep=TRUE; /* Fake out upper level */
		free_pkt(pbuff); /* Free up some memory */
		vlc=new_vc();
		if (!vlc) return; /* Can't even clear the call, sorry */
	}
	vlc->lcn=lci;
	vlc->parent=link;

	insert_lcn(vlc); /* Add to his parent's list */
	set_p(vlc,rec->state,0); /* Either P3 or P5, else a NOOP */

	if (keep || (spclev() == CRITICAL) ) /* Forget it, no resources */ {
		set_p(vlc,P6,Congestion+120);
		return;
	}
	vlc->peer=lcp=new_vc();
	if (lcp == NULL) /* No memory */ {
		set_p(vlc,P6,Congestion+120);
		return;
	}
	lcp->peer=vlc;
	i=(bgetch(&pbuff) & 0x0ff);	/* Now point to x.121 addresses */
	adrstr[0]=i;			/* need to save this */
	i= (i & 0x0f) + (i >> 4);
	i= ( (i & 1) + i) >> 1;
	for (j=1;j<=i;j++) {
		if ( (ch=bgetch(&pbuff))==EOF) break;
		adrstr[j]=ch; /* Copy the address field */
	}
	if (ch != EOF) {
		adr_copy(0,adrstr,0,lcp->addr);
		adr_copy(1,adrstr,0,vlc->addr);
		j=bgetch(&pbuff); /* Get length of copy */
		str_work[0]=j;
		for (i=1;i<=j;i++) str_work[i]=bgetch(&pbuff);
		i=x25_facilities(vlc,str_work);
		if (i>=0) /* Insure the required facilities included */
		  if (!(vlc->nsap[0] &&
		  	lcp->nsap[0] &&
		  	lcp->addr[0])) i= -2;
	} else i= -3;
	if (i<0) /* Call failed */ {
		free_vc(vlc->peer);
		set_p(vlc,P6,Invalid_Fac+76);
		return;
	}
	i=0;
	while (i<=32) /* Max user data field */ {
		ch=bgetch(&pbuff);
		if (ch==EOF) {
			vlc->data[0]=lcp->data[0]=i;
			break;
		}
		else {
			i++;
			if (i<33) vlc->data[i]=lcp->data[i]=ch;
		}
	}
	if (ch != EOF) /* Did not hit EOF ! */ {
		free_vc(vlc->peer);
		set_p(vlc,P6,Invalid_Fac+39);
		return;
	}
	if (vlc->rand == 0) /* First switch to get the call */ {
		if (call_num == 0) call_num++;
		vlc->rand = lcp->rand = (call_num<<8) + rand();
		call_num++;
	}
	else /* Check to see if we have seen this call before! */ {
		get_nsap(vlc,&vlc_adr);
		get_nsap(lcp,&lcp_adr);
		lnk=links;
		while (lnk) /* Check all links */ {
			vc=lnk->llcn;
			while (vc) /* Check all vc's */ {
				if (vc->P == P2) /* Outgoing call? */ {
					get_nsap(vc,&ax_adr1);
					get_nsap(vc->peer,&ax_adr2);
					if ( (vc->rand == vlc->rand)  &&
					(callcmp(&ax_adr1,&lcp_adr)) &&
					(callcmp(&ax_adr2,&vlc_adr)) ) {
					  /* It all Matches! Routing Failure! */
						free_vc(vlc->peer);
						set_p(vlc,P6,Not_Obtainable+120);
						return;
					}
				}
				vc=vc->next;
			}
			lnk=lnk->next;
		}
	}
	lcp->alt = 255;	/* First time through call_request */
	call_request(lcp);
}
