/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "COSI Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "COSI Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "COSI 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */
struct pt_str {
        int     external;
        int     internal; } PT []={
                -1,0,   /* Undefined or Unsupported */
                0,P_DATA,
                1,P_RR,
                5,P_RNR,
                9,0,            /* unsupported P_REJECT, */
                27,P_RS,
                31,P_RSC,
                35,P_INTR,
                39,P_INTRC,
                -2,P_SHORT,
                -7,P_LCNNOT0,
                11,P_INCALL,
                15,P_CONNECTED,
                19,P_CLR,
                23,P_CLRC,
                -6,P_LCN0,
                241,P_DIAGNOSTIC,
                251,P_RST,
                255,P_RSTC,
/*              243,P_REG, */  243,0, /* NOT SUPPORTED */
/*              247,P_REGC */  247,0  /* NOT SUPPORTED */
        };

#define CONN524 pdiscard
#define CONN525 pdiscard

int rstate_call();
int r_error();
int rdiscard();

int pstcall();
int p_error();
int pdiscard();

int dstcall();
int d_error();
int ddiscard();

int diagn();
int diagnostic();

int rsrt42();
int rsrt43();
int rsrt44();

int incall();
int clear554();
int clear552();

int data();
int rr_rnr();

int intr2();
int intrc3();

struct table_entry {
        int (*action)();
        char state;
        char diag; };

struct table_entry trst[21][3]={ /* 21 x 3 x 3 */
    { { pstcall,0,0},	/* Invalid Packet Type */
      { rdiscard,0,0},
      { r_error,R2,33} },

    { { pstcall,0,0},	/* Data Packet */
      { rdiscard,R2,0},
      { r_error,R2,19} },

    { { pstcall,0,0},	/* RR Packet */
      { rdiscard,R2,0},
      { r_error,R2,19} },

    { { pstcall,0,0},	/* RNR Packet */
      { rdiscard,R2,0},
      { r_error,R2,19} },

    { { pstcall,0,0},	/* Reject Packet (Unsupported) */
      { rdiscard,R2,0},
      { r_error,R2,19} },

    { { pstcall,0,0},	/* Reset Packet */
      { rdiscard,R2,0},
      { r_error,R2,19} },

    { { pstcall,0,0},	/* Reset Confirm Packet */
      { rdiscard,R2,0},
      { r_error,R2,19} },

    { { pstcall,0,0},	/* Interrupt Packet */
      { rdiscard,R2,0},
      { r_error,R2,19} },

    { { pstcall,0,0},	/* Interrupt Confirm Packet */
      { rdiscard,R2,0},
      { r_error,R2,19} },

    { { pstcall,0,0},	/* Packet too short */
      { rdiscard,R2,0},
      { r_error,R2,38} },

    { { pstcall,0,0},	/* LCN not 0 when should be */
      { rdiscard,R2,0},
      { r_error,R2,41} },

    { { pstcall,0,0},	/* Incoming Call (aka Call Request) */
      { rdiscard,R2,0},
      { r_error,R2,19} },

    { { pstcall,0,0},	/* Call Connected (aka Call Accept) */
      { rdiscard,R2,0},
      { r_error,R2,19} },

    { { pstcall,0,0},	/* Clear Indication (aka Clear Request) */
      { rdiscard,R2,0},
      { r_error,R2,19} },

    { { pstcall,0,0},	/* Clear Confirm */
      { rdiscard,R2,0},
      { r_error,R2,19} },

    { { diagn,R1,36},	/* LCN 0 when it shouldn't be */
      { diagn,R2,36},
      { diagn,R3,36} },

    { { diagnostic,R1,0}, /* Diagnostic Packet */
      { diagnostic,R2,0},
      { diagnostic,R3,0} },

    { { rsrt42,R3,0},	/* Packet Level Restart Request */
      { rsrt43,R1,0}, /* REALLY GOTO P1, ACTION ROUTINE WILL HANDLE */
      { rdiscard,R3,0} },

    { { r_error,R2,17},	/* Restart Confirm */
      { rsrt44,R1,0}, /* REALLY P1, ACTION ROUTINE WILL HANDLE */
      { r_error,R2,19} },

    { { diagn,R1,0}, /* registration */
      { diagn,R2,0}, /* registration */
      { diagn,R3,0} }, /* registration */

    { { diagn,R1,0}, /* registrationc */
      { diagn,R2,0}, /* registrationc */
      { diagn,R3,0} }  /* registrationc */
};

struct table_entry pst[15][7]={
    { { p_error,P6,33},	/* Invalid Packet Type */
      { p_error,P6,33},
      { p_error,P6,33},
      { dstcall,0,0},
      { p_error,P6,33},
      { pdiscard,P6,0},
      { p_error,P6,33} },

    { { p_error,P6,20},	/* Data Packet */
      { p_error,P6,21},
      { p_error,P6,22},
      { dstcall,0,0},
      { p_error,P6,24},
      { pdiscard,P6,0},
      { p_error,P6,26} },

    { { p_error,P6,20},	/* RR Packet */
      { p_error,P6,21},
      { p_error,P6,22},
      { dstcall,0,0},
      { p_error,P6,24},
      { pdiscard,P6,0},
      { p_error,P6,26} },

    { { p_error,P6,20},	/* RNR Packet */
      { p_error,P6,21},
      { p_error,P6,22},
      { dstcall,0,0},
      { p_error,P6,24},
      { pdiscard,P6,0},
      { p_error,P6,26} },

    { { p_error,P6,20},	/* Reject Packet (Unsupported) */
      { p_error,P6,21},
      { p_error,P6,22},
      { dstcall,0,0},
      { p_error,P6,24},
      { pdiscard,P6,0},
      { p_error,P6,26} },

    { { p_error,P6,20},	/* Reset Packet */
      { p_error,P6,21},
      { p_error,P6,22},
      { dstcall,0,0},
      { p_error,P6,24},
      { pdiscard,P6,0},
      { p_error,P6,26} },

    { { p_error,P6,20},	/* Reset Confirm Packet */
      { p_error,P6,21},
      { p_error,P6,22},
      { dstcall,0,0},
      { p_error,P6,24},
      { pdiscard,P6,0},
      { p_error,P6,26} },

    { { p_error,P6,20},	/* Interrupt Packet */
      { p_error,P6,21},
      { p_error,P6,22},
      { dstcall,0,0},
      { p_error,P6,24},
      { pdiscard,P6,0},
      { p_error,P6,26} },

    { { p_error,P6,38},	/* Interrupt Confirm Packet */
      { p_error,P6,38},
      { p_error,P6,38},
      { dstcall,0,0},
      { p_error,P6,38},
      { pdiscard,P6,0},
      { p_error,P6,38} },

    { { p_error,P6,41},	/* Packet too short */
      { p_error,P6,41},
      { p_error,P6,41},
      { dstcall,0,0},
      { p_error,P6,41},
      { pdiscard,P6,0},
      { p_error,P6,41} },

    { { p_error,P6,41},	/* LCN not 0 when should be */
      { p_error,P6,41},
      { p_error,P6,41},
      { dstcall,0,0},
      { p_error,P6,41},
      { pdiscard,P6,0},
      { p_error,P6,41} },

    { { incall,P3,0},	/* Incoming Call (aka Call Request) */
      { incall,P5,0},
      { p_error,P6,22},
      { p_error,P6,23},
      { p_error,P6,24},
      { pdiscard,P6,0},
      { p_error,P6,26} },

    { { p_error,P6,20},	/* Call Connected (aka Call Accept) */
      { CONN524,P4,0},
      { p_error,P6,22},
      { p_error,P6,23},
      { CONN525,P4,0},
      { pdiscard,P6,0},
      { p_error,P6,26} },

    { { clear552,P7,0},	/* Clear Indication (aka Clear Request) */
      { clear552,P7,0},
      { clear552,P7,0},
      { clear552,P7,0},
      { clear552,P7,0},
      { clear554,P1,0},
      { pdiscard,P7,0} },

    { { p_error,P6,20},	/* Clear Confirmation */
      { p_error,P6,21},
      { p_error,P6,22},
      { p_error,P6,23},
      { p_error,P6,24},
      { clear554,P1,0},
      { p_error,P6,26} }
};

struct table_entry dst[11][3]={
    { { d_error,D2,33},	/* Invalid Packet Type */
      { ddiscard,D2,0},
      { d_error,D2,33} },

    { { data,0,0},	/* Data Packet */
      { ddiscard,D2,0},
      { d_error,D2,29} },

    { { rr_rnr,0,0},	/* RR Packet */
      { ddiscard,D2,0},
      { d_error,D2,29} },

    { { rr_rnr,0,0},	/* RNR Packet */
      { ddiscard,D2,0},
      { d_error,D2,29} },

    { { diagn,D2,33},	/* Reject Packet (Unsupported) */
      { ddiscard,D2,0},	/* should never happen, disabled in table */
      { d_error,D2,29} },

    { { ddiscard,D3,0},   /* The set_d() handles reset 82 */
      { ddiscard,D1,0},   /* The set_d() handles reset 83 */
      { ddiscard,D3,0} },

    { { d_error,D2,27},	/* Reset Confirm Packet */
      { ddiscard,D1,0},   /* The set_d() handles reset 84 */
      { d_error,D2,29} },

    { { intr2,D1,0},	/* Interrupt Packet */
      { ddiscard,D2,0},
      { d_error,D2,29} },

    { { intrc3,D1,0},	/* Interrupt Confirm Packet */
      { ddiscard,D2,0},
      { d_error,D2,29} },

    { { d_error,D2,38},	/* Packet too short */
      { ddiscard,D2,0},
      { d_error,D2,38} },

    { { d_error,D2,41},	/* LCN not 0 when should be */
      { ddiscard,D2,0},
      { d_error,D2,41} }
};


