/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */
#include "ptypes.h"
#include "tx.h"

#include <conio.h>
#undef NULL

#include "timer.h"
#include "data.h"
#include "buffer.h"
#include "ax25.h"
#include "ax25l2.h"
#include "l3struc.h"
#include "iface.h"

#include "x25cause.h"
#include "l3tables.d"

#define adcl(s,d,p) adcsgn(s,d,p,&L3CSIZ,L3SIZ)

#include <malloc.h>

static int  gfi, qbit, dbit, lci, ptype, P_TYPE, mbit, pr, ps, paclen;
static unsigned char keep;

static struct axcb *link;
static struct VCS *lc;
static struct datastr *pbuff;

unsigned char str_work[256];

#include "misc.c"

#include "setstate.c"

#include "l2test.c"

#include "l3calls.c"

#include "prps.c"

#include "tx.c"

#include "dxfer.c"

#include "restart.c"

#include "clear.c"

#include "routing.c"

#include "paths.c"

#include "facility.c"

#include "calsetup.c"

#include "intr.c"

#include "incall.c"

rdiscard(rec)
register struct table_entry *rec;
{
	set_r(link,rec->state,DTE_Orig+rec->diag);
}

pdiscard(rec)
register struct table_entry *rec;
{
	if (lc != NULL)	set_p(lc,rec->state,DTE_Orig+rec->diag);
}

ddiscard(rec)
register struct table_entry *rec;
{
	set_d(lc,rec->state,rec->diag);
}

rstate_call(curr)
int curr;

{
	register struct table_entry *x;

	x= &trst[P_TYPE][curr];
	(*x->action)(x);
}

pstcall(rec)
struct table_entry *rec;
{
	static unsigned char curp;
	register struct table_entry *x;

	if (lc==NULL) curp=P1;
	else curp=lc->P;

	x= &pst[P_TYPE][curp];
	(*x->action)(x);
}

dstcall(rec)
struct table_entry *rec;
{
	register struct table_entry *x;

	x= &dst[P_TYPE][lc->D];
	(*x->action)(x);
}

r_error(rec)
register struct table_entry *rec;
{
	set_r(link,rec->state,DTE_Orig+rec->diag);
}

p_error(rec)
struct table_entry *rec;
{
	register struct VCS *vc;

	vc=lc;
	if (vc == NULL) /* Allocate a vc to send a clear */ {
		vc=new_vc();
		if (vc != NULL) /* Just in case */ {
			vc->parent=link;
			vc->lcn=lci;
			insert_lcn(vc);
		}
	}
	if (vc != NULL) set_p(vc,P6,DTE_Orig+rec->diag);
}

d_error(rec)
struct table_entry *rec;
{
	set_d(lc,D2,rec->diag);
}

diagn(rec)
register struct table_entry *rec;
{
	if (link != NULL) tx_diag(link,rec->diag);
	set_r(link,rec->state,DTE_Orig+rec->diag);
}

diagnostic(rec)
struct table_entry *rec;
{
/*	printf(" Diagnostic Received: %d\n",bgetch(&pbuff)); */
}

void
level3(lkn)	/* This is the r_upcall for Level 3 Links */
struct axcb *lkn;
{

	static int  i;
	int recv_ax25();

	link=lkn;	 /* Get Length and Data */
	if ((paclen = recv_ax25(link,&pbuff,0)) < 0) return;

	keep=FALSE;
	if (paclen < 2) P_TYPE= -38 ;
	else {
	        gfi=bgetch(&pbuff);
	        qbit=gfi & QBIT;
	        dbit=gfi & DBIT;
	        lci=((gfi & 0x0f)<<8) + bgetch(&pbuff);
	        gfi=gfi>>4;
	        ptype = -2;
	        pr=0;
	        ps=0;
	        mbit=0;
	        if (paclen>2) {
	                ptype=bgetch(&pbuff);
	                if (ptype!=0xf1) { /* Diagnostic Packet */
	                        pr=(ptype&0xe0)>>5;
	                        if ((ptype & 1)==0) { /* Info Packet */
	                                mbit=(ptype&MBIT);
	                                ps=(ptype&0x0e)>>1;
	                                ptype=0; }
	                        else if ((ptype & 2)==0) ptype=(ptype&0x1f);
	                                else pr=0;
	                 }
	        }
	        P_TYPE=0; /* default to invalid */
	        for (i=0;i<PT_MAX;i++) {
	                if (PT[i].external==ptype) {
	                        P_TYPE=PT[i].internal;
	                        break;
	                        }
	                }
		lc=NULL;
		if ((lci>254) || (lci>link->maxvc)) P_TYPE= -36;
		else {
			if ((lci != 0) && (link->R == R1)) /* Has a VC */ {
				lc=link->llcn;
			        while (lc && lc->lcn!=lci) lc=lc->next;
		        }
			if (P_TYPE >= 0) {
	        	        if (P_TYPE == P_INCALL) i=0x0b;
		                else if (P_TYPE == P_DATA) i=0x03;
		                        else i=0x0f;
		                if( (gfi & i) != 1) P_TYPE= -40;
		                else if (P_TYPE >= P_DIAGNOSTIC) {
	        	                if (lci != 0) P_TYPE = P_LCNNOT0; }
	                	     else
	                        	if (lci==0) P_TYPE = P_LCN0;
		        }
		}
	}
	if (P_TYPE < 0) tx_diag(link, -P_TYPE);
	else {
		i=link->R;
		if (i != R0) i--;
	        rstate_call(i);
	}
	if (!keep) free_pkt(pbuff);
}
