/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */
extern unsigned char str_work[256];

set_d(vc,new_d,diag)
register struct VCS *vc;
unsigned char new_d;
unsigned char diag;

{
	static struct VCS *vcp;

	puthex2("Set D",new_d,"For",vc->parent);

	vcp=vc->peer;

	if (new_d != vc->D)
		switch (new_d) {
		case D1:
			vc->LIPR=vc->OPR=vc->OPS=vc->NIPS=
				vc->intr=vc->LBUSY=vc->RBUSY=0;
			kill(vc->timer);
			vc->D=D1;
			free_queue(&vc->tx_queue);
			break;
		case D2:
			(*vc->RESET)(vc,diag);
		case D3: /* NOTE FALL THROUGH */
			kill(vc->timer);
			vc->D=new_d;
			if (vcp && vcp->P==P4 && vcp->D==D1) set_d(vcp,D2,diag);
			else {
				set_p(vc,P6,Local_Proc);
				break;
			}
			if (new_d == D3) {
				tx_crst(vc);
				set_d(vc, D1, 0);
			}
			break;
		}
}

set_p(vc,new_p,causex)
register struct VCS *vc;
unsigned char new_p;
unsigned int causex;

{
	int addlcn();
	static unsigned char i;
	static int cause;
	static struct VCS *vcx, *vcp;

	cause=causex;
	vcp=vc->peer;

	puthex2("Set P",new_p,"For",vc->parent);

	if ((new_p == P1) || (vc->P != new_p))
	switch(new_p) {
		case P1: /* Init and discard */
			vc->P=new_p;
			set_d(vc,D1,0);
			dellcn(vc);
			if (vcp) vcp->peer=NULL;
			vc->peer=NULL;
			vcx=vc;
			free_vc(vcx);
			break;
		case P2:
			if ((addlcn(vc) == TRUE)) {
				if (tx_callreq(vc) == EOF)
					set_p(vc,P6,Congestion+127);
				else vc->P = P2;
			}
			else /* Failed */ {
				if (vcp) {
					vcp->peer=NULL;
					set_p(vcp,P6,Congestion+71);
				}
				set_p(vc,P1,0);
			}
			break;
		case P3:
		case P5:
			vc->P=new_p;
			break;
		case P4:
			i=0;
			if (!vcp) /* NULL */ i=64;
			else
				switch (vc->P) {
					case P2:
						vc->P=P4;
						set_p(vcp,P4,cause);
						break;
					case P3:
						vc->P=P4;
						if (tx_accept(vc) == EOF)
						    set_p(vc,P6,Congestion+127);
						break;
					case P5:
						vc->P=P4;
						if (vc->parent->am_dce)
							set_p(vcp,P4,cause);
						else {
						  if (tx_accept(vcp) == EOF)
						    set_p(vc,P6,Congestion+127);
						}
						break;
					default:
						i=vc->P + 20;
						break;
				}
			if (i) set_p(vc,P6,Local_Proc+i);
			else {
				set_d(vc,D1,0);
				(*vc->CONNECTED)(vc);
			}
			break;
		case P6:
			vc->P=P6;
			if (vcp) /* Not NIL */
				if (vcp->P < P6) set_p(vcp,P6,cause);
			(*vc->CLEAR)(vc,cause);
			break;
		case P7:
			vc->P=P7;
			if (vcp) /* not NULL */ {
				if (vcp->lcn == 255) /* a pending call */ {
				 	dellcn(vcp);
					set_p(vcp, P1,0);
					tx_cclr(vc);
					set_p(vc,P1,0);
				}
				else set_p(vcp,P6,cause);
			}
			else {
				tx_cclr(vc);
				set_p(vc,P1,0);
			}
			break;
	}
}

set_r(lnk,new_r,c)
register struct axcb *lnk;
int new_r,c;

{
	puthex2("Set R", new_r,"For",lnk);

	if (lnk->R != new_r) {
		switch (new_r) {
			case R0:
				lnk->R=R0;
				break;
			case R1:
				lnk->R=R1;
				if ((lnk->link_num < 16) && (lnk->link_num>0))
					route_success(lnk); /* Network Node */
				else
					process_calls(lnk); /* User L3 Link */
				break;
			case R2:
				lnk->R=R2;
				tx_rsrt(lnk, c);
				break;
			case R3:
				lnk->R=R3;
				tx_crsrt(lnk);
				break;
		}
	}
}

