/*
 * Copyright 1988 by the Radio Amateur Telecommunications Society
 * and Thomas A. Moulton, W2VY
 *
 * This software may only be modified, copied, distributed or
 * executed for non-profit purposes by individuals operating
 * systems in the Amateur Radio Service.  Credit to the
 * author(s) and to the Radio Amateur Telecommunications Society
 * must be made in modules where RATS provided software is used,
 * and in any announcements and documentation.  
 *
 * As a non-profit, research and development organization,  the
 * Radio Amateur Telecommunications Society distributes software
 * in both executable and source forms.  This policy is in place
 * to encourage the development and distribution of OSI-based,
 * networking tools.  In order to protect the interests of the
 * Society and the authors, we have placed some conditions
 * of use on the software.  Other groups are encouraged
 * to place the same or similar guidelines on
 * software they produce.
 *
 * The Radio Amateur Telecommunications Society reserves the right
 * to specify and alter the terms under which software provided by
 * the Society may be used.  This policy is consistent with the 
 * objective of uniform and consistent "Open Systems Interconnections."
 * 
 * All acceptable Amateur Radio related uses of this software
 * will be outlined in the "ROSE Implementer's Guide".  Individuals
 * or organizations wishing to add to, or modify the provisions of
 * the guide to accommodate local or evolutionary requirements
 * should document the proposed change(s) and forward them to the
 * Society.  If accepted, written notification will be provided by
 * the Society to the submitting organization or individual(s).
 * The Society will then issue a "ROSE Implementer's Guide Change
 * Notice".  Periodically, the Society will re-issue the "ROSE 
 * Implementer's Guide" and incorporate the text of the change 
 * notices.  This procedure has been put in to place to ensure
 * compatibility between systems and to ensure their "Openness"
 * and interoperability.
 *
 * No part of this software may be used in other packages 
 * without prior authorization from the author or the Society.  
 * Software incorporating this module, all or in part, must be 
 * provided to the Society prior to distribution or use by
 * anyone not directly involved in testing of the revised  
 * environment.  Current releases of the combined software must
 * be provided to the Society in both source and executable
 * forms.  Adequate documention to produce an executable module 
 * from the provided source must also be included.
 *
 * Non-Amateur Radio non-profit uses may be authorized on a case
 * by case basis.  Inquiries for such use may be made in writing
 * to the Society. Non-commercial uses consistent with the
 * general principles of Open Systems Interconnection Reference
 * Model will be generally considered with favor.
 *
 * Commercial licensing of the software is also available based
 * on normal commercial terms.  Licensing inquiries should be
 * directed to the Society.  Commercial licensing of the standard
 * software will be done in situations which materially benefit
 * the Amateur Radio Packet Network.  Additional licensing is
 * reserved by the individual authors.
 * 
 * The Radio Amateur Telecommunications Society provides this software
 * on an "as is" basis.  The Society assumes no liability for
 * loss incurred through the use of this software.  Amateur Radio
 * use of this software implies non-commercial and voluntary 
 * development, deployment and use of this software in a "Amateur",
 * non-commercial service.  Commercial users are encouraged to
 * inspect their copies of the source code.  Source code modification
 * licenses are available if a combined Object and Source Code
 * license was not originally established.
 * 
 * The Society may be contacted by writing or calling at:
 * 
 * The Radio Amateur Telecommunications Society 
 * 206 North Vivyen Street.
 * Bergenfield, New Jersey 07621
 *
 * Telephone: 201-387-8896
 *
 */
/* This file contains the routines that create packets
   and sends them to Level 2 */

/* These routines all assume a LCGN of 0, which limits the number of
   calls on a given link to 1-255, but the real max LCN # is 254 with
   255 used to indicate that the VC is a pending call, 0 is used to indicate
   a Level 2 (AX.25) User */

struct datastr *get_pkt(vc, size)
register struct VCS *vc;
int size;
{
	static struct datastr *p;

	if ( (p=new_buffer(size)) == NULL) return NULL;
	bappch(p,0x10); /* GFI and LCGN=0 */
	if (vc != NULL) /* This is not a supervisory packet */
		bappch(p,vc->lcn);
	else /* Supervisory */
		bappch(p,0); /* LCN=0 */
	return p;
}

int
tx_pkt0(lnk, vc, pcode)
struct axcb *lnk;
struct VCS *vc;
unsigned char pcode;

{
	register struct datastr *pkt;

	pkt=get_pkt(vc,3);
	if (pkt == NULL) return EOF; /* No Memory! */
	bappch(pkt,pcode); /* Packet Type */
	l2_tx(lnk,pkt);
	return 0;
}

int
tx_pkt2(lnk, vc, pcode, c)
struct axcb *lnk;
struct VCS *vc;
unsigned int pcode, c;

{
	register struct datastr *pkt;

	pkt=get_pkt(vc,5);
	if (pkt == NULL) return EOF;	/* No Memory! */
	bappch(pkt,pcode); /* Packet Type */
	bappch(pkt,(c>>8)); /* Cause */
	bappch(pkt,(c&0x0ff)); /* Diagnostic */
	l2_tx(lnk,pkt);
	return 0;
}

int
tx_diag(lnk,d)
struct axcb *lnk;
unsigned char d;

{
	register struct datastr *pkt;

	pkt=get_pkt(NULL,4);
	if (pkt == NULL) return EOF;
	bappch(pkt,0xf1); /* Diagnostic */
	bappch(pkt,d); /* Diagnostic Code/Reason */
	l2_tx(lnk,pkt);
	return 0;
}

int
tx_intr(vc,pktx)
struct VCS *vc;
struct datastr *pktx;
{
	register  struct datastr *pkt1;
	static struct datastr *pkt;
	static int ch;

	pkt1=new_buffer(buflen(pkt=pktx));
	if (!pkt1) return EOF;		/* No Memory, We are dead */
	bappch(pkt,0x10); /* GFI =01, LCGN=0 */
	bappch(pkt,vc->lcn);
	while ( (ch=bgetch(&pkt)) != EOF) bappch(pkt1,ch);
	l2_tx(vc->parent,pkt1);
	return 0;
}

int
tx_callreq(vcx)
struct VCS *vcx;
{
	register struct datastr *pkt;
	static struct VCS *vc;

	pkt=get_pkt((vc=vcx),128);	/* Biggest call request packet */
	if (pkt == NULL) return EOF;	/* No Memory! */
	bappch(pkt,0x0b); /* Call Request */
	put_fac(vc,pkt);
	l2_tx(vc->parent,pkt);
	return 0;
}

void
tx_rr(vc)
register struct VCS *vc;
{
	static int i;

	if (vc->P == P4 && vc->D == D1) {
		queue(&vc->timer,tx_rr,60,vc);	/* Reset timer, if running */
		i=(vc->OPR<<5)+1;	/* Assume RR */
		if (vc->LBUSY) i += 4;	/* Send RNR */
		tx_pkt0(vc->parent,vc,i);
	}
}
